# -*- coding: utf-8 -*-
import sys
import redfish

# When running on the server locally use the following commented values
# iLO_host = "blobstore://."
# LOGIN_ACCOUNT = "None"
# LOGIN_PASSWORD = "None"

# When running remotely connect using the iLO address, iLO account name, 
# and password to send https requests
iLO_host = "https://10.0.0.100"
login_account = "admin"
login_password = "password"

## Create a REDFISH object
REDFISH_OBJ = redfish.RedfishClient(base_url=iLO_host,username=login_account, \
                          password=login_password)

# Login into the server and create a session
REDFISH_OBJ.login()

# Do a GET on a given path
response = REDFISH_OBJ.get("/redfish/v1/systems/1")

# Print out the response
sys.stdout.write("%s\n" % response)

# Logout of the current session
REDFISH_OBJ.logout()


# -*- coding: utf-8 -*-
"""
Expanded LegacyREST/Redfish interface for schema validation, database for responses, caching,
and error registries.
"""

from .ris import (
    RisInstanceNotFoundError,
    RisMonolith,
    RisMonolithMemberBase,
    RisMonolithMemberv100,
    SessionExpired,
)
from .rmc import RmcApp
from .rmc_helper import (
    CurrentlyLoggedInError,
    IdTokenError,
    IloLicenseError,
    InstanceNotFoundError,
    InvalidSelectionError,
    NothingSelectedError,
    NothingSelectedFilterError,
    NothingSelectedSetError,
    RmcCacheManager,
    RmcFileCacheManager,
    ScepenabledError,
    UndefinedClientError,
    ValidationError,
    ValueChangedError,
)
from .sharedtypes import JSONEncoder
from .validation import RegistryValidationError, ValidationManager



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of gathering the smart array encryption settings on HPE iLO systems
"""

import sys
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def get_SmartArray_EncryptionSettings(_redfishobj, desired_properties):

    smartstorage_response = []
    smartarraycontrollers = dict()

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        systems_uri = _redfishobj.root.obj['Systems']['@odata.id']
        systems_response = _redfishobj.get(systems_uri)
        systems_members_uri = next(iter(systems_response.obj['Members']))['@odata.id']
        systems_members_response = _redfishobj.get(systems_members_uri)
        smart_storage_uri = systems_members_response.obj.Oem.Hpe.Links\
                                                                ['SmartStorage']['@odata.id']
        smart_storage_arraycontrollers_uri = _redfishobj.get(smart_storage_uri).obj.Links\
                                                                ['ArrayControllers']['@odata.id']
        smartstorage_response = _redfishobj.get(smart_storage_arraycontrollers_uri).obj['Members']
    else:
        for instance in resource_instances:
            #Use Resource directory to find the relevant URI
            if '#HpeSmartStorageArrayControllerCollection.' in instance['@odata.type']:
                smartstorage_uri = instance['@odata.id']
                smartstorage_response = _redfishobj.get(smartstorage_uri).obj['Members']
                break

    for controller in smartstorage_response:
        smartarraycontrollers[controller['@odata.id']] = _redfishobj.get(controller['@odata.id']).\
                                                                                                obj
        sys.stdout.write("Encryption Properties for Smart Storage Array Controller \'%s\' : \n" \
                                        % smartarraycontrollers[controller['@odata.id']].get('Id'))
        for data in smartarraycontrollers[controller['@odata.id']]:
            if data in desired_properties:
                sys.stdout.write("\t %s : %s\n" % (data, smartarraycontrollers[controller\
                                                                        ['@odata.id']].get(data)))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #list of desired properties related to Smart Array controller encryption
    DESIRED_PROPERTIES = ["Name", "Model", "SerialNumber", "EncryptionBootPasswordSet",\
             "EncryptionCryptoOfficerPasswordSet",\
             "EncryptionLocalKeyCacheEnabled", "EncryptionMixedVolumesEnabled",\
             "EncryptionPhysicalDriveCount", "EncryptionRecoveryParamsSet",\
             "EncryptionStandaloneModeEnabled", "EncryptionUserPasswordSet"]
    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    get_SmartArray_EncryptionSettings(REDFISHOBJ, DESIRED_PROPERTIES)
    REDFISHOBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of testing connection to the ESKM system for HPE iLO systems
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def test_ESKM_connection(_redfishobj):

    eskm_uri = None
    body = dict()

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        managers_uri = _redfishobj.root.obj['Managers']['@odata.id']
        managers_response = _redfishobj.get(managers_uri)
        managers_members_uri = next(iter(managers_response.obj['Members']))['@odata.id']
        managers_members_response = _redfishobj.get(managers_members_uri)
        security_service_uri = managers_members_response.obj.Oem.Hpe.Links['SecurityService']\
                                                                                    ['@odata.id']
        security_service_response = _redfishobj.get(security_service_uri)
        eskm_uri = security_service_response.obj.Links['ESKM']['@odata.id']

    else:
        #Use Resource directory to find the relevant URI
        for instance in resource_instances:
            if '#HpeESKM.' in instance['@odata.type']:
                eskm_uri = instance['@odata.id']
                break

    if eskm_uri:
        eskm_test_conn_uri = _redfishobj.get(eskm_uri).obj['Actions']\
                                                    ['#HpeESKM.TestESKMConnections']['target']
        body["Action"] = "HpeESKM.TestESKMConnections"
        resp = _redfishobj.post(eskm_test_conn_uri, body)
        #If iLO responds with something outside of 200 or 201 then lets check the iLO extended info
        #error message to see what went wrong
        if resp.status == 400:
            try:
                print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, \
                                                                                sort_keys=True))
            except Exception as excp:
                sys.stderr.write("A response error occurred, unable to access iLO Extended "\
                                 "Message Info...")
        elif resp.status != 200:
            sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
        else:
            print("Success!\n")
            print(json.dumps(resp.dict, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    test_ESKM_connection(REDFISHOBJ)
    REDFISHOBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of setting the Manager ethernet interface static IP
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def set_ilo_static_ipv4(_redfishobj, ipv4_dict, dns_dict):

    ethernet_data = {}
    body = dict()

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        managers_uri = _redfishobj.root.obj['Managers']['@odata.id']
        managers_response = _redfishobj.get(managers_uri)
        managers_members_uri = next(iter(managers_response.obj['Members']))['@odata.id']
        managers_members_response = _redfishobj.get(managers_members_uri)
        manager_ethernet_interfaces = managers_members_response.obj['EthernetInterfaces']\
                                                                                    ['@odata.id']
        manager_ethernet_interfaces_response = _redfishobj.get(manager_ethernet_interfaces)
        manager_ethernet_interfaces_members = manager_ethernet_interfaces_response.\
                                                            obj['Members']
        for _member in manager_ethernet_interfaces_members:
            _tmp = _redfishobj.get(_member['@odata.id']).obj
            ethernet_data[_member['@odata.id']] = _tmp
    else:
        #Use Resource directory to find the relevant URI
        for instance in resource_instances:
            if '#EthernetInterfaceCollection.' in instance['@odata.type'] and 'Managers' in \
                                                                        instance['@odata.id']:
                ethernet_uri = instance['@odata.id']
                ethernet_interfaces = _redfishobj.get(ethernet_uri).obj['Members']
                for _ethernet_interface in ethernet_interfaces:
                    ethernet_data[_ethernet_interface['@odata.id']] = _redfishobj.\
                                                        get(_ethernet_interface['@odata.id']).dict
                break

    if ethernet_data:
        print("\n\nShowing all available ethernet management interfaces before changes:\n\n")
        for interface in ethernet_data:
            sys.stdout.write("Ethernet Management Inteface \'%s\'\n" % ethernet_data\
                                                                            [interface].get('Id'))

            sys.stdout.write("\'DHCPv4\':\n")
            if ethernet_data[interface].get('DHCPv4'):
                print(json.dumps(ethernet_data[interface].get('DHCPv4'), indent=4, sort_keys=True))
            else:
                print(json.dumps(ethernet_data[interface]['Oem']['Hpe'].get('DHCPv4'), indent=4, \
                                                                                sort_keys=True))
            sys.stdout.write("\'IPv4\':\n")
            if ethernet_data[interface].get('IPv4StaticAddresses'):
                print(json.dumps(ethernet_data[interface].get('IPv4Addresses'), indent=4, \
                                                                                sort_keys=True))
            sys.stdout.write("\'StaticNameServers\':\n")
            if ethernet_data[interface].get('StaticNameServers'):
                print(json.dumps(ethernet_data[interface].get('StaticNameServers'), indent=4, \
                                                                                sort_keys=True))

    for ethernet in ethernet_data:
        sys.stdout.write("Ethernet Interface: %s\n" % ethernet)
        ans = input("Would you like to modify this interface? (y/n)\n")
        if "n" in ans:
            continue
        if 'DHCPv4' in ethernet_data[ethernet]:
            if ethernet_data[ethernet]['DHCPv4'].get('UseDNSServers'):
                resp = _redfishobj.patch(ethernet, {"DHCPv4": {"UseDNSServers": False}})
                ilo_response(_redfishobj, resp)
            if ethernet_data[ethernet]['DHCPv4'].get('UseGateway'):
                resp = _redfishobj.pathc(ethernet, {"DHCPv4": {"UseGateway": False}})
                ilo_response(_redfishobj, resp)
        if 'IPv4StaticAddresses' in ethernet_data[ethernet]:
            body.update({"IPv4Addresses": [ipv4_dict]})
        if 'StaticNameServers' in ethernet_data[ethernet]:
            body.update({"StaticNameServers" : [dns_dict.get('PrimaryDNS'), \
                                                                    dns_dict.get('SecondaryDNS')]})
        else:
            body.update({"Oem": {"Hpe": {"IPv4": {"DNSServers": [dns_dict.get('PrimaryDNS'), \
                                                                 dns_dict.get('SecondaryDNS')]}}}})
        resp = _redfishobj.patch(ethernet, body)
        ilo_response(_redfishobj, resp)
        break

def ilo_response(_redfishobj, resp):
    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, \
                                                                            sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended "\
                             "Message Info...")
    elif resp.status != 200:
        sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
    else:
        print("Success! Suggest to reset iLO for settings to take effect.\n")
        print(json.dumps(resp.dict, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #IPv4 settings for Address, Gateway and SubnetMask as well as DNS.
    IPV4_DICT = {'Address':'16.83.61.123', \
                 'Gateway':'16.83.60.1', \
                 'SubnetMask':'255.255.252.0' \
                 }
    DNS_DICT = {'PrimaryDNS':'16.110.135.52', \
                'SecondaryDNS':'16.110.135.51' \
                }

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    set_ilo_static_ipv4(REDFISHOBJ, IPV4_DICT, DNS_DICT)
    REDFISHOBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-

"""A convenience layer that combines multiple lower level classes and functions into one."""

import copy
import hashlib
import logging
import shutil

# ---------Imports---------
import sys
import time

import six

try:
    from collections import OrderedDict
except ImportError:
    from collections.abc import OrderedDict

import jsonpatch
import jsonpointer

import redfish.ris.gen_compat
import redfish.ris.validation
from redfish.rest.v1 import RestClient
from redfish.ris.resp_handler import ResponseHandler
from redfish.ris.ris import RisMonolith, SchemaValidationError, SessionExpired
from redfish.ris.rmc_helper import (
    EmptyRaiseForEAFP,
    IloResponseError,
    IncompatibleiLOVersionError,
    InstanceNotFoundError,
    LoadSkipSettingError,
    NothingSelectedError,
    NothingSelectedSetError,
    RmcFileCacheManager,
    UndefinedClientError,
    ValidationError,
    ValueChangedError,
)
from redfish.ris.utils import (
    checkallowablevalues,
    diffdict,
    getattributeregistry,
    iterateandclear,
    merge_dict,
    navigatejson,
    print_handler,
    skipnonsettingsinst,
    validate_headers,
    warning_handler,
)
from redfish.ris.validation import Typepathforval, ValidationManager

# ---------End of imports---------

# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------


class RmcApp(object):
    """A convenience class that combines the client, compatibility, validation, caching,
    and monolith into one class.

    :param showwarnings: Flag to print warnings to std.out (True) or log in log file (False)
    :type showwarnings: bool
    :param cache_dir: Cache directory to save cache data to, if None, RmcApp will not cache data.
                      Cache can allow your RmcApp to persist between scripts.
    :type cache_dir: str
    """

    def __init__(self, showwarnings=False, cache_dir=None):
        self.logger = LOGGER
        self.redfishinst = None
        self.sessioninst = None
        self._cm = RmcFileCacheManager(self)
        self.monolith = None
        self._iloversion = None
        self._validationmanager = None
        self._selector = None
        self._cachedir = cache_dir
        self.verbose = 1
        self._sessionid = None
        self.typepath = redfish.ris.gen_compat.Typesandpathdefines()
        Typepathforval(typepathobj=self.typepath)

        if not showwarnings:
            self.logger.setLevel(logging.WARNING)
            if self.logger.handlers and self.logger.handlers[0].name == "lerr":
                self.logger.handlers.remove(self.logger.handlers[0])

    @property
    def monolith(self):
        """Get the monolith from the current client"""
        return self._monolith

    @monolith.setter
    def monolith(self, monolith):
        """Set the monolith"""
        self._monolith = monolith

    @property
    def current_client(self):
        """Get the current client"""
        if self.redfishinst:
            return self.redfishinst
        raise UndefinedClientError()

    @property
    def validationmanager(self):
        """Get the valdation manager"""
        if self.getiloversion():
            if self._validationmanager:
                self._validationmanager.reset_errors_warnings()
            else:
                monolith = self.monolith
                self._validationmanager = ValidationManager(monolith, defines=self.typepath)
            self._validationmanager.updatevalidationdata()
        else:
            self._validationmanager = None
        return self._validationmanager

    @property
    def selector(self):
        """The selector that will be used to gather data if no selector or instance argument is
        passed."""
        return self._selector

    @selector.setter
    def selector(self, sel):
        """Set the selector"""
        self._selector = sel

    @property
    def cachedir(self):
        """The cache directory that is used to cache app data to a file,
        None if not caching data."""
        return self._cachedir

    @cachedir.setter
    def cachedir(self, cache_dir):
        """Set the cachedir"""
        self._cachedir = cache_dir

    @property
    def cache(self):
        """True if we are caching data, False if we are not"""
        return True if self.cachedir else False

    def restore(self, creds=None, enc=False):
        """Restores the monolith from cache. Used to load a monolith data back into a new app
        class. Keyword arguments are only needed in a local client when in a high security mode.

        :param creds: Credentials to create the client with.
        :type creds: str
        :param enc: Flag to determine if encoding functions are being used. True if being used false
                    if not.
        :type enc: bool
        """
        self._cm.uncache_rmc(creds=creds, enc=enc)

    def set_encode_funct(self, funct):
        """Set the encoding function for cache to use. Can be used to protect sensitive data when
        it is at rest.

        :param funct: The function to use for encoding data
        :type funct: function
        """
        self._cm.encodefunct = funct

    def set_decode_funct(self, funct):
        """Set the decoding function for cache to use. Is used in conjunction with the
        `set_encode_funct` to turn the encoded data back into a usable string.

        :param funct: The function to use for decoding data
        :type funct: function
        """
        self._cm.decodefunct = funct

    def save(self):
        """Updates the cache with the latest monolith data."""
        self._cm.cache_rmc()

    def login(
        self,
        username=None,
        password=None,
        sessionid=None,
        base_url="blobstore://.",
        path=None,
        skipbuild=False,
        includelogs=False,
        biospassword=None,
        is_redfish=False,
        proxy=None,
        ssl_cert=None,
        user_ca_cert_data=None,
        json_out=False,
        login_otp=None,
        log_dir=None,
    ):
        """Performs a login on a the server specified by the keyword arguments. Will also create
        a monolith, client, and update the compatibility classes for the app instance. If base_url
        is not included the login is assumed to be locally on the OS.

        :param username: The user name required to login to server.
        :type: str
        :param password: The password credentials required to login.
        :type password: str
        :param base_url: The redfish host name or ip address to login to.
        :type base_url: str
        :param path: The path to initiate the monolith crawl from. If None, it will start from the
                     root. See monolith documentation on how the path is used.
        :type path: str
        :param proxy: The proxy required for connection (if any).
        :type proxy: str
        :param ssl_cert: The path to the CA bundle or SSL certificate to use with connection
                        (if any).
        :type ssl_cert: str
        :param user_ca_cert_data: Dictionary of user certificate data for iLO Certificate-based
          authentication including iLO User TLS certificate, iLO User CA Root Key, \
          iLO User CA Root Key Password (for encrypted CAs)
        :type: user_ca_pass: str
        :param skipbuild: The flag to determine monolith download. If True, monolith will be
                          initiated empty, if False will build the monolith.
        :type skipbuild: bool
        :param includelogs: The flag to determine id logs should be downloaded in the crawl.
        :type includelogs: bool
        :param biospassword: The BIOS password for the server if set.
        :type biospassword: str
        :param is_redfish: If True, a Redfish specific header (OData) will be
            added to every request. Only required if the system has both LegacyREST and Redfish.
        :type is_redfish: bool
        """

        self.typepath.getgen(
            url=base_url,
            username=username,
            password=password,
            sessionid=sessionid,
            ca_cert_data=user_ca_cert_data,
            proxy=proxy,
            isredfish=is_redfish,
            login_otp=login_otp,
            log_dir=log_dir,
        )
        if user_ca_cert_data and self.typepath.iloversion < 5.23:
            raise IncompatibleiLOVersionError(
                "Certificate based login is incompatible with this " "iLO version: %s\n" % self.typepath.iloversion
            )
        is_redfish = self.typepath.updatedefinesflag(redfishflag=is_redfish)

        if self.redfishinst and self.redfishinst.session_key:
            self.logout()

        self.redfishinst = RestClient(
            base_url=base_url,
            username=username,
            password=password,
            session_key=sessionid,
            default_prefix=self.typepath.defs.startpath,
            biospassword=biospassword,
            is_redfish=is_redfish,
            proxy=proxy,
            ca_cert_data=user_ca_cert_data,
            login_otp=login_otp,
            log_dir=log_dir,
        )

        self.current_client.login(self.current_client.auth_type)

        inittime = time.time()

        self._build_monolith(path=path, includelogs=includelogs, skipbuild=skipbuild, json_out=json_out)
        endtime = time.time()

        if self.verbose > 1:
            sys.stdout.write("Monolith build process time: %s\n" % (endtime - inittime))
        self.save()
        if not self.monolith:
            self.monolith.update_member(
                resp=self.current_client.root,
                path=self.typepath.defs.startpath,
                init=False,
            )

    def logout(self, url=None):
        """Performs a logout of the server and prepares the app for another system, setting app
        variables to default values.

        :param url: The URL for the logout request. Only needed when using a cache.
        :type url: str
        """
        sessionlocs = []
        self._validationmanager = None
        self._iloversion = None

        try:
            self.monolith.killthreads()
        except Exception:
            pass

        try:
            self.current_client.logout()
        except Exception:
            sessionlocs = self._cm.logout_del_function(url)
        else:
            self._cm.logout_del_function(url)

        for session in sessionlocs:
            try:
                self.delete_handler(session[0], silent=True, service=True)
            except:
                pass
        self.redfishinst = None

        cachedir = self.cachedir
        if cachedir:
            try:
                shutil.rmtree(cachedir)
            except Exception:
                pass

    def select(self, selector=None, fltrvals=(None, None), path_refresh=False):
        """Selects instances based on selector and filter values. The select specified is saved in
        the app for further use. If another selector is sent, it overwrites the current one.

        :param selector: The type (@odata.type for Redfish) to select.
        :type selector: str
        :param fltrvals: The filter values for the select operation (Key,Val). If a selector returns
                         multiple instances fltrvals can filter the instances by a key/value pair,
                         limiting the returned instances to the one you want.
        :type fltrvals: tuple
        :param path_refresh: The flag to reload the selected instances. If True, each instance will be
                    grabbed again from the server to make sure responses are up to date.
        :type path_refresh: bool
        :returns: A list of selected monolith member instances.
        :rtype: RisMonolithMemberv100
        """
        if not selector:
            selector = self.selector
        # Still nothing selected and selector is NULL
        if not selector:
            raise NothingSelectedError()
        selector = self.typepath.modifyselectorforgen(selector)
        instances = self._getinstances(selector=selector, path_refresh=path_refresh)
        val = fltrvals[1].strip("'\"") if isinstance(fltrvals[1], six.string_types) else fltrvals[1]
        instances = [
            inst
            for inst in instances
            if not fltrvals[0] or navigatejson(fltrvals[0].split("/"), copy.deepcopy(inst.dict), val)
        ]
        if any(instances):
            self.selector = selector
            self.save()
            return instances

        errmsg = (
            "Unable to locate instance for '{0}' and filter '{1}={2}'".format(selector, fltrvals[0], fltrvals[1])
            if fltrvals[0] and fltrvals[1]
            else "Unable to locate instance for {}\n".format(selector)
        )

        raise InstanceNotFoundError(errmsg)

    def types(self, fulltypes=False):
        """Returns a list of types available to be queried and selected with monolith.

        :param fulltypes: Flag to determine if types return Redfish full name, if False will return
                          a shortened version of the type string.
        :type fulltypes: bool
        :returns: A list of type strings.
        :rtype: list
        """
        instances = list()

        if self.monolith:
            monolith = self.monolith
            rdirtype = next(monolith.gettypename(self.typepath.defs.resourcedirectorytype), None)

            if not rdirtype:
                for inst in monolith.iter():
                    if not any([x for x in ["ExtendedError", "object", "string"] if x in inst.type]):
                        instances.append(inst.type)
            else:
                for instance in monolith.iter(rdirtype):
                    for item in instance.resp.dict["Instances"]:
                        if (
                            item
                            and instance._typestring in list(item.keys())
                            and "ExtendedError" not in item[instance._typestring]
                        ):
                            if not fulltypes and instance._typestring == "@odata.type":
                                tval = item["@odata.type"].split("#")
                                tval = tval[-1].split(".")[:-1]
                                tval = ".".join(tval)
                                instances.append(tval)
                            elif item:
                                instances.append(item[instance._typestring])
        return instances

    def getprops(
        self,
        selector=None,
        props=None,
        nocontent=None,
        skipnonsetting=True,
        remread=False,
        insts=None,
    ):
        """Gets properties from a specified selector. If no selector is specified, uses the selector
        property in the app class. Instead of a selector a list of instances to search can be used
        instead. If both **selector** and **insts** are passed, **insts** is used.

        Specific values for multi-level dictionaries can be returned by passing each key
        separated by a "/" Ex: Key/Sub-Key/Sub-Sub-Key

        :param selector: The type selection for the get operation.
        :type selector: str
        :param skipnonsetting: Flag to remove non settings path.
        :type skipnonsetting: bool
        :param nocontent: Keys not found are added to the list provided.
        :type nocontent: list
        :param remread: Flag to remove readonly properties.
        :type remread: bool
        :param props: The keys to search for within current selection.
        :type props: list
        :param insts: List of RisMonolithMemberv100 to be searched for specific keys.
        :type insts: list
        :returns: A list of properties found in dictionary form.
        :rtype: list
        """
        results = list()
        nocontent = set() if nocontent is None else nocontent
        if props:
            if isinstance(props, list):
                noprop = {prop: False for prop in props}
            else:
                noprop = {props: False}
        else:
            noprop = {}
        instances = insts if insts else self._getinstances(selector=selector)
        instances = skipnonsettingsinst(instances) if skipnonsetting else instances

        if selector != "UpdateService.":
            if not instances or len(instances) == 0:
                if not selector:
                    raise NothingSelectedError()

        for instance in instances:
            currdict = instance.dict
            for patch in instance.patches:
                currdict = jsonpatch.apply_patch(currdict, patch)
            _ = self.removereadonlyprops(currdict, emptyraise=True) if remread else None
            temp_dict = dict()
            if props:
                if isinstance(props, six.string_types):
                    props = [props]
                for prop in props:
                    copydict = copy.deepcopy(currdict)
                    propsdict = navigatejson(prop.split("/"), copydict)
                    if propsdict is None:
                        continue
                    noprop[prop] = True
                    merge_dict(temp_dict, propsdict)
                if temp_dict:
                    if temp_dict not in results:
                        results.append(temp_dict)
            else:
                results.append(currdict)
        if props:
            _ = [nocontent.add(prop) for prop in props if not noprop[prop]]
        return results

    def info(self, selector=None, props=None, dumpjson=True, latestschema=False):
        """Gets schema information for properties from a specified selector. If no selector is
        specified, uses the selector property in the app class. If no properties are specified
        the entire schema dictionary is returned in a list.

        :param selector: The type selection for the info operation.
        :type selector: str
        :param props: The keys to gather schema data for within current selection.
        :type props: str or list
        :param dumpjson: Flag to determine if output should be human readable or json schema.
        :type dumpjson: bool
        :param latestschema: Flag to determine if we should drop the schema version when we try to
                             match schema information. If True, the version will be dropped.
        :type latestschema: bool
        :returns: A list of property schema information if dumpjson is True or string if dumpjson is
                  False.
        :rtype: list or string
        """
        model = None
        outdata = ""
        nokey = False
        results = None
        typestring = self.typepath.defs.typestring
        iloversion = self.getiloversion()
        if not iloversion:
            return results
        instances = self._getinstances(selector)
        attributeregistry = getattributeregistry(instances)
        instances = skipnonsettingsinst(instances)

        if not instances or len(instances) == 0:
            raise NothingSelectedError()

        for inst in instances:
            bsmodel = None
            currdict = inst.resp.dict
            proppath = inst.resp.getheader("Link").split(";")[0].strip("<>") if inst.resp.getheader("Link") else None
            seldict = {}
            if not props:
                model, bsmodel = self.get_model(currdict, attributeregistry, latestschema, proppath=proppath)
                results = model
                break
            if isinstance(props, six.string_types):
                props = props.split("/") if "/" in props else props
                props = [props] if not isinstance(props, (list, tuple)) else props
                seldict = navigatejson(props, copy.deepcopy(currdict))
                if seldict is None:
                    nokey = True
                    continue
            if self.typepath.defs.typestring in currdict:
                seldict[typestring] = currdict[typestring]
                model, bsmodel = self.get_model(
                    currdict,
                    attributeregistry,
                    latestschema,
                    newarg=props[:-1],
                    proppath=proppath,
                )
            if not model and not bsmodel:
                errmsg = "/".join(props)
                warning_handler(
                    "Unable to locate registry model or " "No data available for entry: {}\n".format(errmsg)
                )
                continue
            found = model.get_validator(props[-1]) if model else None
            found = bsmodel.get_validator(props[-1]) if not found and bsmodel else found
            outdata = found if found and dumpjson else found.print_help(props[-1]) if found else outdata

        if outdata or results:
            return outdata if outdata else results

        errmsg = (
            "Entry {} not found in current selection\n".format("/".join(props))
            if nokey
            else "Entry {} not found in current" " selection\n".format("/".join(props))
        )
        warning_handler(errmsg)

    def loadset(
        self,
        seldict=None,
        fltrvals=(None, None),
        diffonly=False,
        latestschema=False,
        uniqueoverride=False,
        selector=None,
    ):
        """Creates json patches in monolith if the supplied dictionary passes schema validation.
        In the event schemas are unavailable the patches are always added. Patches that are created
        this way are not sent to the server until the :meth:`commit` function is called, sending the
        patches to the server. A list of patches that have not been sent to the server can be
        returned with the :meth:`status` function.

        :param selector: The type selection for the loadset operation.
        :type selector: str
        :param seldict: Dictionary with the patches to apply to the selected instances.
        :type seldict: dict
        :param fltrvals: The filter values for the operation (Key,Val). If a selector returns
                         multiple instances fltrvals can filter the instances by a key/value pair,
                         limiting the returned instances to the one you want. If no filter is
                         supplied the patch dictionary will be applied to all instances.
        :type fltrvals: tuple
        :param latestschema: Flag to determine if we should drop the schema version when we try to
                             match schema information. If True, the version will be dropped.
        :type latestschema: bool
        :param diffonly: flag to differentiate only existing properties.
        :type diffonly: bool
        :param uniqueoverride: Flag to determine if system unique properties should also be patched.
                               If this is True, then unique properties will be patched.
        :type uniqueoverride: bool
        :returns: returns a list of properties that have successfully been set
        """
        results = list()
        nochangesmade = False
        settingskipped = [False]

        selector = self.selector if not selector else selector
        instances = self.select(selector=selector, fltrvals=fltrvals)
        attributeregistry = getattributeregistry(instances=instances)
        instances = skipnonsettingsinst(instances=instances)

        if not instances or len(instances) == 0:
            raise NothingSelectedSetError()

        for instance in instances:
            if validate_headers(instance, verbose=self.verbose):
                continue
            else:
                nochangesmade = True

            currdict = instance.resp.dict
            if "@odata.id" in seldict:
                if currdict["@odata.id"] != seldict["@odata.id"]:
                    continue
            diff_resp = diffdict(
                newdict=copy.deepcopy(seldict),
                oridict=copy.deepcopy(currdict),
                settingskipped=settingskipped,
            )

            iloversion = self.getiloversion()
            if iloversion:
                proppath = (
                    instance.resp.getheader("Link").split(";")[0].strip("<>")
                    if instance.resp.getheader("Link")
                    else None
                )
                try:
                    self._validatechanges(
                        instance=instance,
                        attributeregistry=attributeregistry,
                        newdict=diff_resp,
                        oridict=currdict,
                        unique=uniqueoverride,
                        latestschema=latestschema,
                        proppath=proppath,
                    )
                except SchemaValidationError:
                    LOGGER.error("Cannot validate changes, error found in schema.")

            patches = jsonpatch.make_patch(currdict, diff_resp)

            if patches:
                torem = []
                _ = [torem.append(patch) for patch in patches.patch if patch["op"] == "remove"]
                _ = [patches.patch.remove(patch) for patch in torem]

            for ind, item in enumerate(instance.patches):
                ppath = item.patch[0]["path"] if hasattr(item, "patch") else item[0]["path"]
                # ppath = ["path"](getattr(item, "patch"), item)[0]["path"]
                jpath = jsonpointer.JsonPointer(ppath.lower())
                jval = jpath.resolve(seldict, default="kasjdk?!")
                if not jval == "kasjdk?!":
                    del instance.patches[ind]

            if patches:
                for patch in patches.patch:
                    forprint = patch["value"] if "value" in patch else (patch["op"] + " " + patch["from"])
                    results.append({patch["path"][1:]: forprint})
                if "Managers/1/EthernetInterfaces/1" not in instance.path:
                    self.monolith.path(instance.path).patches.append(patches)
            else:
                nochangesmade = True

        if not nochangesmade:
            return results
        elif settingskipped[0] is True:
            raise LoadSkipSettingError()
        else:
            return results

    def status(self):
        """Returns all pending changes that have not been committed yet."""
        iloversion = self.getiloversion()

        finalresults = list()
        monolith = self.monolith
        (_, _) = self.get_selection(setenable=True)
        attrreg = getattributeregistry([ele for ele in monolith.iter() if ele])
        for instance in monolith.iter():
            results = list()

            if not (instance.patches and len(instance.patches) > 0):
                continue
            for item in instance.patches:
                if isinstance(item, list):
                    results.extend(jsonpatch.JsonPatch(item))
                else:
                    results.extend(item)

            currdict = instance.resp.dict
            itemholder = list()
            for mainitem in results:
                item = copy.deepcopy(mainitem)

                if iloversion:
                    _, bsmodel = self.get_model(currdict, attrreg)
                    if bsmodel:
                        prop = item["path"][1:].split("/")[-1]
                        validator = bsmodel.get_validator(prop)
                        if validator:
                            if isinstance(validator, redfish.ris.validation.PasswordValidator):
                                item["value"] = "******"

                itemholder.append(item)
            if itemholder:
                finalresults.append({instance.maj_type + "(" + instance.path + ")": itemholder})

        return finalresults

    def commit(self):
        """Applies all pending json patches to the server.

        :yields: Two strings.

                1. Path being PATCHed
                2. True if an error occurred during the PATCH, False if no error.
        """
        # Protect iLO Network Interface changes.
        instances = [
            inst for inst in self.monolith.iter() if inst.patches and "Managers/1/EthernetInterfaces/1" not in inst.path
        ]

        if not instances or len(instances) == 0:
            raise NothingSelectedError()

        for instance in instances:
            if validate_headers(instance, verbose=self.verbose):
                continue

            currdict = dict()
            oridict = instance.resp.dict
            totpayload = dict()
            # apply patches to represent current edits
            for patches in instance.patches:
                if not self._iloversion:
                    self._iloversion = self.getiloversion()
                if self._iloversion < 5.130:
                    self._checkforetagchange(instance=instance)
                fulldict = jsonpatch.apply_patch(oridict, patches)
                for patch in patches:
                    currdict = copy.deepcopy(fulldict)
                    patchpath = patch["path"]
                    pobj = jsonpointer.JsonPointer(patchpath)
                    indpayloadcount = 0
                    for item in pobj.parts:
                        payload = pobj.walk(currdict, item)
                        indpayloadcount = indpayloadcount + 1
                        if isinstance(payload, list):
                            break
                        else:
                            if not isinstance(payload, dict):
                                break
                            currdict = copy.deepcopy(payload)
                    indices = pobj.parts[:indpayloadcount]
                    createdict = lambda x, y: {x: y}
                    while len(indices):
                        payload = createdict(indices.pop(), payload)
                    merge_dict(totpayload, payload)
                currdict = copy.deepcopy(totpayload)
            if "PersistentBootConfigOrder" in currdict and type(currdict) is dict:
                refdict = list(set(currdict["PersistentBootConfigOrder"]))
                items = list(currdict["PersistentBootConfigOrder"])
                items_to_keep = items[: int(len(refdict))]
                currdict = {"PersistentBootConfigOrder": items_to_keep}
            if currdict:
                yield instance.resp.request.path

                put_path = instance.resp.request.path
                etag = self.monolith.paths[put_path].etag
                headers = dict([("If-Match", etag)]) if self._iloversion > 5.130 else None
                for key, value in currdict.items():
                    if key == "OldAdminPassword" and (value != ""):
                        if (value == "null") or (value == "none") or (value == "None"):
                            currdict["OldAdminPassword"] = ""
                        else:
                            hash = hashlib.sha256(value.encode()).hexdigest().upper()
                            headers = dict([("X-HPRESTFULAPI-AuthToken", hash)])
                    elif key == "AdminPassword" and (
                        (value is None) or (value == "none") or (value == "None") or (value == "null")
                    ):
                        currdict["AdminPassword"] = ""
                try:
                    self.patch_handler(
                        put_path,
                        currdict,
                        optionalpassword=self.current_client.bios_password,
                        headers=headers,
                    )
                except IloResponseError:
                    yield True  # Failure
                else:
                    yield False  # Success

    def patch_handler(
        self,
        put_path,
        body,
        headers=None,
        silent=False,
        service=False,
        optionalpassword=None,
    ):
        """Performs the client HTTP PATCH operation with monolith and response handling support.
        Response handling will output to logger or string depending on showmessages app argument.

        :param put_path: The REST path to perform the patch operation on.
        :type put_path: str
        :param body: the body to perform the operation with.
        :type body: dict
        :param headers: Any additional headers to be added to the request.
        :type headers: dict
        :param optionalpassword: The bios password if it is required for the operation.
        :type optionalpassword: str
        :param silent: If False response will be parsed based on service flag and output to a log or
                       stdout. If True response will not be parsed and no message output or error
                       messages raised from the response handler.
        :type silent: bool
        :param service: When handling the response, if True registries will be gathered and a full,
                        response will be output if False they will not and response handler will
                        instead return a generic message.
        :type service: bool
        :returns: A :class:`redfish.rest.containers.RestResponse` object containing response data
        """
        (put_path, body) = self._checkpostpatch(body=body, path=put_path, patch=True)

        if optionalpassword:
            self.current_client.bios_password = optionalpassword

        results = self.current_client.patch(put_path, body=body, headers=headers)

        if results and getattr(results, "status", None) and results.status == 401:
            raise SessionExpired()

        self._modifiedpath(results, replace=True)

        # if results and getattr(results, "status", None) and results.status == 412:
        if results and hasattr(results, "status") and results.status == 412:
            self._updatemono(path=put_path, path_refresh=True)

        if not silent and hasattr(self.typepath.defs, "messageregistrytype"):
            ResponseHandler(self.validationmanager, self.typepath.defs.messageregistrytype).output_resp(
                results, dl_reg=service, verbosity=self.verbose
            )

        return results

    def get_handler(
        self,
        get_path,
        sessionid=None,
        silent=False,
        uncache=False,
        headers=None,
        service=False,
        username=None,
        password=None,
        base_url=None,
    ):
        """Performs the client HTTP GET operation with monolith and response handling support.
        Response handling will output to logger or string depending on showmessages app argument.

        :param put_path: The REST path to perform the get operation on.
        :type put_path: str
        :param uncache: flag to not store the data downloaded into monolith.
        :type uncache: bool
        :param headers: Any additional headers to be added to the request.
        :type headers: dict
        :param silent: If False response will be parsed based on service flag and output to a log or
                       stdout. If True response will not be parsed and no message output or error
                       messages raised from the response handler.
        :type silent: bool
        :param service: When handling the response, if True registries will be gathered and a full,
                        response will be output if False they will not and response handler will
                        instead return a generic message.
        :type service: bool
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        try:
            results = self.current_client.get(get_path, headers=headers)
        except UndefinedClientError:
            if sessionid:
                self.redfishinst = RestClient(
                    sessionid=sessionid,
                    username=username,
                    password=password,
                    base_url=base_url,
                )
                if not headers:
                    headers = dict()
                headers["X-Auth-Token"] = sessionid
                results = self.current_client.get(get_path, headers=headers)

        if results and getattr(results, "status", None) and results.status == 404:
            if not silent:
                if hasattr(self.typepath.defs, "messageregistrytype"):
                    ResponseHandler(self.validationmanager, self.typepath.defs.messageregistrytype).output_resp(
                        results, dl_reg=service, verbosity=self.verbose
                    )
                else:
                    print_handler("[" + str(results.status) + "]" + " The operation completed successfully.\n")
            return results

        if results and results.status == 200 and sessionid:
            if not silent:
                if hasattr(self.typepath.defs, "messageregistrytype"):
                    ResponseHandler(self.validationmanager, self.typepath.defs.messageregistrytype).output_resp(
                        results, dl_reg=service, verbosity=self.verbose
                    )
                else:
                    print_handler("[" + str(results.status) + "]" + " The operation completed successfully.\n")
            return results

        if not uncache and results.status == 200 and not sessionid:
            if self.monolith:
                self.monolith.update_member(resp=results, path=get_path, init=False)

        if results and getattr(results, "status", None) and results.status == 401:
            raise SessionExpired()

        if not silent:
            if hasattr(self.typepath.defs, "messageregistrytype"):
                ResponseHandler(self.validationmanager, self.typepath.defs.messageregistrytype).output_resp(
                    results, dl_reg=service, verbosity=self.verbose
                )
            else:
                print_handler("[" + str(results.status) + "]" + " The operation completed successfully.\n")

        return results

    def post_handler(self, put_path, body, headers=None, silent=False, service=False):
        """Performs the client HTTP POST operation with monolith and response handling support.
        Response handling will output to logger or string depending on showmessages app argument.

        :param put_path: The REST path to perform the post operation on.
        :type put_path: str
        :param body: the body to perform the operation with.
        :type body: dict
        :param headers: Any additional headers to be added to the request.
        :type headers: dict
        :param silent: If False response will be parsed based on service flag and output to a log or
                       stdout. If True response will not be parsed and no message output or error
                       messages raised from the response handler.
        :type silent: bool
        :param service: When handling the response, if True registries will be gathered and a full,
                        response will be output if False they will not and response handler will
                        instead return a generic message.
        :type service: bool
        :returns: A :class:`redfish.rest.containers.RestResponse` object containing response data
        """
        (put_path, body) = self._checkpostpatch(body=body, path=put_path)

        results = self.current_client.post(put_path, body=body, headers=headers)

        if results and getattr(results, "status", None) and results.status == 401:
            raise SessionExpired()

        self._modifiedpath(results)

        if results.status == 400 and results.dict is None:
            return results
        if not silent and hasattr(self.typepath.defs, "messageregistrytype"):
            ResponseHandler(self.validationmanager, self.typepath.defs.messageregistrytype).output_resp(
                results, dl_reg=service, verbosity=self.verbose
            )

        return results

    def put_handler(
        self,
        put_path,
        body,
        headers=None,
        silent=False,
        optionalpassword=None,
        service=False,
    ):
        """Performs the client HTTP PUT operation with monolith and response handling support.
        Response handling will output to logger or string depending on showmessages app argument.

        :param put_path: The REST path to perform the put operation on.
        :type put_path: str
        :param body: the body to perform the operation with.
        :type body: dict
        :param headers: Any additional headers to be added to the request.
        :type headers: dict
        :param optionalpassword: The bios password if it is required for the operation.
        :type optionalpassword: str
        :param silent: If False response will be parsed based on service flag and output to a log or
                       stdout. If True response will not be parsed and no message output or error
                       messages raised from the response handler.
        :type silent: bool
        :param service: When handling the response, if True registries will be gathered and a full,
                        response will be output if False they will not and response handler will
                        instead return a generic message.
        :type service: bool
        :returns: A :class:`redfish.rest.containers.RestResponse` object containing response data
        """
        if optionalpassword:
            self.current_client.bios_password = optionalpassword
        results = self.current_client.put(put_path, body=body, headers=headers)
        if results and getattr(results, "status", None) and results.status == 401:
            raise SessionExpired()

        self._modifiedpath(results, replace=True)

        if not silent and hasattr(self.typepath.defs, "messageregistrytype"):
            ResponseHandler(self.validationmanager, self.typepath.defs.messageregistrytype).output_resp(
                results, dl_reg=service, verbosity=self.verbose
            )

        return results

    def delete_handler(self, put_path, headers=None, silent=False, service=False):
        """Performs the client HTTP DELETE operation with monolith and response handling support.
        Response handling will output to logger or string depending on showmessages app argument.

        :param put_path: The REST path to perform the delete operation on.
        :type put_path: str
        :param headers: Any additional headers to be added to the request.
        :type headers: dict
        :param silent: If False response will be parsed based on service flag and output to a log or
                       stdout. If True response will not be parsed and no message output or error
                       messages raised from the response handler.
        :type silent: bool
        :param service: When handling the response, if True registries will be gathered and a full,
                        response will be output if False they will not and response handler will
                        instead return a generic message.
        :type service: bool
        :returns: A :class:`redfish.rest.containers.RestResponse` object containing response data
        """
        results = self.current_client.delete(put_path, headers=headers)

        if results and getattr(results, "status", None) and results.status == 401:
            raise SessionExpired()
        self._modifiedpath(results, delete=True)

        if not silent and hasattr(self.typepath.defs, "messageregistrytype"):
            ResponseHandler(self.validationmanager, self.typepath.defs.messageregistrytype).output_resp(
                results, dl_reg=service, verbosity=self.verbose
            )

        return results

    def head_handler(self, put_path, silent=False, service=False):
        """Performs the client HTTP HEAD operation with monolith and response handling support.
        Response handling will output to logger or string depending on showmessages app argument.

        :param put_path: The REST path to perform the head operation on.
        :type put_path: str
        :param silent: If False response will be parsed based on service flag and output to a log or
                       stdout. If True response will not be parsed and no message output or error
                       messages raised from the response handler.
        :type silent: bool
        :param service: When handling the response, if True registries will be gathered and a full,
                        response will be output if False they will not and response handler will
                        instead return a generic message.
        :type service: bool
        :returns: A :class:`redfish.rest.containers.RestResponse` object containing response data
        """
        results = self.current_client.head(put_path)

        if results and getattr(results, "status", None) and results.status == 401:
            raise SessionExpired()

        if not silent and hasattr(self.typepath.defs, "messageregistrytype"):
            ResponseHandler(self.validationmanager, self.typepath.defs.messageregistrytype).output_resp(
                results, dl_reg=service, verbosity=self.verbose
            )

        return results

    def removereadonlyprops(self, currdict, emptyraise=False, removeunique=True, specify_props=None):
        """Remove read only properties from a dictionary. Requires schemas to be available.

        :param currdict: The dictionary to remove read only properties from.
        :type currdict: dictionary
        :param emptyraise: Flag to raise an empty error for handling and failure to parse.
        :type emptyraise: boolean
        :type removeunique: Flag to remove system unique values as well as read only.
        :type removeunique: boolean
        :parm specify_props: Optionally set list of properties to be removed instead of the default.
        :type specify_props: list
        """
        try:
            type_str = self.typepath.defs.typestring
            currtype = currdict.get(type_str, None)
            oridict = copy.deepcopy(currdict)
            if specify_props:
                templist = specify_props
            else:
                templist = [
                    "Modified",
                    "Type",
                    "Description",
                    "Status",
                    "links",
                    "SettingsResult",
                    "Attributes",
                    "@odata.context",
                    "@odata.type",
                    "@odata.id",
                    "@odata.etag",
                    "Links",
                    "Actions",
                    "AvailableActions",
                    "BiosVersion",
                    "AddressOrigin",
                ]
            # Attributes removed and readded later as a validation workaround
            currdict = iterateandclear(currdict, templist)
            iloversion = self.getiloversion()
            if not iloversion:
                return currdict
            self.validationmanager.validatedict(
                currdict,
                currtype=currtype,
                monolith=self.monolith,
                unique=removeunique,
                searchtype=None,
            )
            if oridict.get("Attributes", None):
                currdict["Attributes"] = oridict["Attributes"]
            return currdict
        except:
            if emptyraise is True:
                raise EmptyRaiseForEAFP()
            elif emptyraise == "pass":
                pass
            else:
                raise

    def getidbytype(self, tpe):
        """Return a list of URIs that correspond to the supplied type string.

        :param tpe: type string to search for.
        :type tpe: string.
        """
        urls = list()
        val = next(self.monolith.gettypename(tpe), None)
        urls.extend(self.monolith.typesadded[val] if val else [])
        return urls

    def getcollectionmembers(self, path, fullresp=False):
        """Returns collection/item lists of the provided path.

        :param path: path to return.
        :type path: string.
        :param fullresp: Return full json data instead of only members.
        :type path: bool.
        :returns: list of collection members
        """
        if (
            self.typepath.defs.isgen10
            and hasattr(self.typepath, "gencompany")
            and self.typepath.gencompany
            and "?$expand=." not in path
        ):
            path += "?$expand=." if path.endswith("/") else "/?$expand=."

        members = self.get_handler(path, service=True, silent=True)
        if members and not fullresp:
            try:
                members = members.dict["Members"] if self.typepath.defs.isgen10 else members.dict["Current"]
            except KeyError:
                members = members
        elif fullresp:
            members = [members.dict]

        return members

    def getbiosfamilyandversion(self):
        """Function that returns the current BIOS version information."""
        self._updatemono(currtype="ComputerSystem.", crawl=False)

        try:
            for inst in self.monolith.iter("ComputerSystem."):
                if "Current" in inst.resp.obj["Bios"]:
                    oemjson = inst.resp.obj["Bios"]["Current"]
                    parts = oemjson["VersionString"].split(" ")
                    return parts[0], parts[1][1:]
                else:
                    parts = inst.resp.obj["BiosVersion"].split(" ")
                    return parts[0], parts[1][1:]
        except Exception:
            pass

        return None, None

    def getiloversion(self, skipschemas=False):
        """Function that returns the current iLO version.

        :param skipschemas: flag to determine whether to skip schema download. If False, this will
                            also verify if schemas are available.
        :type skipschemas: bool
        :returns: returns current iLO version
        """
        iloversion = self._iloversion = self._iloversion if self._iloversion else self.typepath.iloversion

        if not iloversion and hasattr(self.redfishinst, "iloversion"):
            iloversion = self._iloversion = self.typepath.iloversion = self.redfishinst.iloversion

        if (
            hasattr(self.typepath, "gencompany")
            and self.typepath.gencompany
            and not self._iloversion
            and not self.typepath.noschemas
        ):
            self.monolith.load(self.typepath.defs.managerpath, crawl=False)
            results = next(iter(self.getprops("Manager.", ["FirmwareVersion", "Firmware"])))

            def quickdrill(_dict, key):
                """function to find key in nested dictionary"""
                return _dict[key]

            while isinstance(results, dict):
                results = quickdrill(results, next(iter(results.keys())))
            iloversionlist = results.replace("v", "").replace(".", "").split(" ")
            iloversion = float(".".join(iloversionlist[1:3]))

            model = self.getprops("Manager.", ["Model"])
            if model:
                if next(iter(model))["Model"] == "iLO CM":
                    # Assume iLO 4 types in Moonshot
                    iloversion = None

            self._iloversion = iloversion
        elif (
            hasattr(self.typepath, "gencompany") and not self.typepath.gencompany
        ):  # Assume schemas are available somewhere in non-hpe redfish
            self._iloversion = iloversion = 4.210

        conf = None if not skipschemas else True
        if not skipschemas:
            if iloversion and iloversion >= 4.210:
                conf = self._verifyschemasdownloaded(self.monolith)
            elif iloversion and iloversion < 4.210:
                warning_handler("Please upgrade to iLO 4 version 2.1 or above for schema support.")
            else:
                warning_handler("Schema support unavailable on the currently logged in system.")

        return iloversion if iloversion and iloversion >= 4.210 and conf else None

    def get_selection(self, selector=None, setenable=False, path_refresh=False):
        """Gathers instances and optionally the attributeregistry based on selector.

        :param selector: The type selection for the get operation.
        :type selector: str.
        :param setenable: Flag to determine if registry should also be returned.
        :type setenable: boolean.
        :param path_refresh: Flag to reload the selected instances.
        :type path_refresh: boolean.
        :returns: returns a list of selected items
        """
        instances = self._getinstances(selector=selector, path_refresh=path_refresh)
        if setenable:
            attributeregistryfound = getattributeregistry(instances=instances)
            instances = skipnonsettingsinst(instances=instances)
            return instances, attributeregistryfound

        return instances

    def create_save_header(self):
        """Adds save file headers to show what server the data came from.

        :param selector: The type selection for the get save operation.
        :type selector: str.
        :param selectignore: Return the save header even if there isn't a selection to add it to.
        :type selectignore: boolean
        :returns: returns an header ordered dictionary
        """
        instances = OrderedDict()
        monolith = self.monolith

        self._updatemono(currtype="ComputerSystem.", crawl=False)
        self._updatemono(currtype=self.typepath.defs.biostype, crawl=False)
        self._updatemono(currtype="Manager.", crawl=False)

        instances["Comments"] = OrderedDict()
        for instance in monolith.iter("ComputerSystem."):
            if instance.resp.obj.get("Manufacturer"):
                instances["Comments"]["Manufacturer"] = instance.resp.obj["Manufacturer"]
            if instance.resp.obj.get("Model"):
                instances["Comments"]["Model"] = instance.resp.obj["Model"]
            try:
                if instance.resp.obj["Oem"][self.typepath.defs.oemhp]["Bios"]["Current"]:
                    oemjson = instance.resp.obj["Oem"][self.typepath.defs.oemhp]["Bios"]["Current"]
                    instances["Comments"]["BIOSFamily"] = oemjson["Family"]
                    instances["Comments"]["BIOSDate"] = oemjson["Date"]
            except KeyError:
                pass
        for instance in monolith.iter(self.typepath.defs.biostype):
            try:
                if getattr(instance.resp.obj, "Attributes", False):
                    if instance.resp.obj["Attributes"].get("SerialNumber"):
                        instances["Comments"]["SerialNumber"] = instance.resp.obj["Attributes"]["SerialNumber"]
                if instance.resp.obj.get("SerialNumber"):
                    instances["Comments"]["SerialNumber"] = instance.resp.obj["SerialNumber"]
            except KeyError:
                pass
        for instance in monolith.iter("Manager."):
            if instance.resp.obj.get("FirmwareVersion"):
                instances["Comments"]["iLOVersion"] = instance.resp.obj["FirmwareVersion"]

        return instances

    def download_path(self, paths, crawl=True, path_refresh=False):
        """Loads paths into the monolith.

        :param paths: list of paths to download
        :type paths: list
        :param path_refresh: Flag to reload the paths or not.
        :type path_refresh: bool.
        :param crawl: Flag to determine if load should traverse found links.
        :type crawl: boolean.
        """
        if not paths:
            return
        try:
            list(
                map(
                    lambda x: self.monolith.load(
                        path=x,
                        init=False,
                        path_refresh=path_refresh,
                        crawl=crawl,
                        includelogs=True,
                    ),
                    paths,
                )
            )
        except Exception as excp:
            try:
                if excp.errno == 10053:
                    raise SessionExpired()
            except:
                raise excp
            else:
                raise excp

    def get_model(self, currdict, attributeregistry, latestschema=None, newarg=None, proppath=None):
        """Returns a model and possibly a bios model for the current instance's schema/registry.
        This model can be used to read schema data and validate patches.

        :param currdict: The dictionary to gather the schema model from.
        :type currdict: dict
        :param attributeregistry: The current systems attribute registry. If not gathering a bios
                                  registry this can be set to None.
        :type attributeregistry: dict
        :param latestschema: Flag to determine if we should drop the schema version when we try to
                             match schema information. If True, the version will be dropped.
        :type latestschema: bool
        :param newargs: List of multi level properties to be gathered.
        :type newargs: list
        :param proppath: The path of the schema you want to validate (from Location header).
        :type proppath: str
        :returns: model and bios model
        """
        type_str = self.typepath.defs.typestring
        bsmodel = None
        valobj = self.validationmanager
        model = valobj.get_registry_model(
            currtype=currdict[type_str],
            newarg=newarg,
            latestschema=latestschema,
            proppath=proppath,
        )
        if not attributeregistry and model:
            return model, bsmodel
        if not model and not attributeregistry:
            LOGGER.warning("Unable to locate registry/schema for %s\n", currdict[type_str])
            return None, None
        attrval = currdict.get("AttributeRegistry", None)
        attrval = list(attributeregistry.values())[0] if not attrval and attributeregistry else attrval
        bsmodel = valobj.get_registry_model(
            currtype=attrval if attrval else currdict[type_str],
            newarg=newarg,
            latestschema=latestschema,
            searchtype=self.typepath.defs.attributeregtype,
        )
        return model, bsmodel

    def _build_monolith(self, path=None, includelogs=False, skipbuild=False, json_out=False):
        """Run through the RIS tree to build monolith

        :param path: path to initiate login to.
        :type path: str.
        :param includelogs: flag to determine id logs should be downloaded.
        :type includelogs: boolean.
        :param skipbuild: if true, skip build of monolith (initialize empty)
        :type skipbuild: True
        """
        self.monolith = RisMonolith(self.current_client, self.typepath)
        if not skipbuild:
            self.monolith.load(path=path, includelogs=includelogs, init=True, json_out=json_out)
        else:
            self.monolith.update_member(
                resp=self.current_client.root,
                path=self.current_client.default_prefix,
                init=False,
            )

    def _modifiedpath(self, results, delete=False, replace=False):
        """Check the path and set the modified flag

        :param delete: Flag to delete the path in the results
        :type delete: bool
        :param replace: Flag to replace the path from the results
        :type replace: bool
        :param results: Response for the path
        :type results: RestResponse
        """
        if not results or results.status not in (200, 201):
            return
        path = results.path
        path = path.split("/Actions")[0] if "Actions" in path else path
        path = path + "/" if self.typepath.defs.isgen10 and path[-1] != "/" else path
        if not replace and path in self.monolith.paths:
            self.monolith.paths[path].modified = True
            _ = self.monolith.markmodified(path)
        if delete and path in self.monolith.paths:
            self.monolith.removepath(path)
        if replace and path in self.monolith.paths:
            self.monolith.paths[path].modified = True
            self.monolith.paths[path].patches = []

    def _checkforchange(self, paths, crawl=True):
        """Check if the given paths have been modified and updates monolith if it has

        :param paths: paths to be checked
        :type paths: list
        """
        (pathtoetag, _) = self._gettypeswithetag()
        mono = self.monolith
        self.download_path(list(paths), crawl=crawl, path_refresh=True)
        etags = [None if path not in mono.paths else mono.paths[path].etag for path in paths]
        sametag = [
            path
            for ind, path in enumerate(paths)
            if path in pathtoetag and path in self.monolith.paths and pathtoetag[path] != etags[ind]
        ]
        for path in sametag:
            self.monolith.paths[path].patches = []
        if sametag:
            LOGGER.warning(
                "The data in the following paths have been updated. " "Recheck the changes made to . %s",
                ",".join([str(path) for path in sametag]),
            )

    def _updatemono(self, currtype=None, path=None, crawl=False, path_refresh=False):
        """Check if type/path exists in current monolith

        :param entrytype: the found entry type.
        :type entrytype: str.
        :param currtype: the current entry type.
        :type currtype: str.
        :param crawl: flag to determine if load should traverse found links.
        :type crawl: boolean.
        """
        monolith = self.monolith
        currtype = None if currtype == '"*"' else currtype
        paths = set()
        if currtype:
            for path, resp in monolith.paths.items():
                if currtype and currtype.lower() not in resp.maj_type.lower():
                    continue
                if path_refresh or not resp:
                    paths.add(path)
                if resp:
                    try:
                        if not resp.dict:
                            raise AttributeError
                    except AttributeError:
                        paths.add(path)
                if resp.modified:
                    paths.add(path)
                    paths.update(monolith.checkmodified(path) if path in monolith.ctree else set())
        elif path:
            if monolith.paths and not list(monolith.paths)[0][-1] == "/":
                path = path[:-1] if path[-1] == "/" else path
            if path_refresh or not monolith.path(path):
                paths.add(path)
            if path in monolith.paths and monolith.paths[path].modified:
                paths.add(path)
                paths.update(monolith.checkmodified(path) if path in monolith.ctree else set())
        if paths:
            self._checkforchange(list(paths), crawl=crawl)

    def _verifyschemasdownloaded(self, monolith):
        """Function to verify that the schema has been downloaded

        :param monolith: full data model retrieved from server.
        :type monolith: dict.
        """

        schemaid = self.typepath.schemapath
        regid = self.typepath.regpath

        if not (schemaid and regid):
            warning_handler("Missing Schemas or registries.")
            return None

        schemacoll = next(monolith.gettypename(self.typepath.defs.schemafilecollectiontype), None)
        if not schemacoll or any(
            paths.lower() == schemaid and monolith.paths[paths] for paths in monolith.typesadded[schemacoll]
        ):
            self.download_path([schemaid], crawl=False)
            schemacoll = next(monolith.gettypename(self.typepath.defs.schemafilecollectiontype), None)

        regcoll = next(monolith.gettypename(self.typepath.defs.regfilecollectiontype), None)
        if not regcoll or any(
            paths.lower() == regid and monolith.paths[paths] for paths in monolith.typesadded[regcoll]
        ):
            self.download_path([regid], crawl=False)
            regcoll = next(monolith.gettypename(self.typepath.defs.regfilecollectiontype), None)

        return any(
            paths.lower() in (schemaid.lower(), regid.lower()) and monolith.paths[paths] for paths in monolith.paths
        )

    def _validatechanges(
        self,
        instance=None,
        attributeregistry=None,
        latestschema=None,
        proppath=None,
        newdict=None,
        oridict=None,
        unique=False,
    ):
        """Validate the changes that are requested by the user.

        :param newdict: dictionary with only the properties that have changed
        :type newdict: dict.
        :param oridict: selection dictionary with current state.
        :type oridict: dict.
        :param unique: flag to determine override for unique properties.
        :type unique: str.
        :param iloversion: current iLO version.
        :type iloversion: float.
        :param instance: current selection instance.
        :type instance: RisMonolithMemberv100.
        :param attrreg: Registry entry of the given attribute.
        :type attrreg: RepoRegistryEntry.
        """
        entrymono = self.monolith
        currtype = oridict[self.typepath.defs.typestring]
        validation_manager = self.validationmanager
        errors, warnings = validation_manager.validatedict(
            newdict,
            currtype=attributeregistry[instance.maj_type] if attributeregistry else currtype,
            monolith=entrymono,
            unique=unique,
            searchtype=self.typepath.defs.attributeregtype if attributeregistry else None,
            latestschema=latestschema,
            proppath=proppath,
        )

        validation_errors = errors
        for warninngs in warnings:
            warning_handler(warninngs, override=True)
        if validation_errors and len(validation_errors) > 0:
            raise ValidationError(validation_errors)
        checkallowablevalues(newdict=newdict, oridict=oridict)

    def _getinstances(self, selector=None, path_refresh=False, crawl=False):
        """Main function to get instances of particular type and reload

        :param selector: the type selection for the get operation.
        :type selector: str.
        :param setenable: flag to determine if registry should also be returned.
        :type setenable: boolean.
        :param setenable: flag to determine if registry should also be returned.
        :type setenable: boolean.
        :param path_refresh: flag to reload the selected instances.
        :type path_refresh: boolean.
        :returns: returns a list of selected items
        """
        instances = list()
        selector = self.selector if not selector else selector
        if selector:
            selector = ".".join(selector.split("#")[-1].split(".")[:2])
            if self.monolith:
                self._updatemono(currtype=selector, crawl=crawl, path_refresh=path_refresh)
        if not selector:
            return instances
        selector = None if selector == '"*"' else selector
        if self.monolith:
            if self.redfishinst.is_redfish:
                instances = [
                    inst
                    for inst in self.monolith.iter(selector)
                    if inst.maj_type not in ["object", "string"] and "redfish" in inst.path
                ]
            else:
                instances = [
                    inst
                    for inst in self.monolith.iter(selector)
                    if inst.maj_type not in ["object", "string"] and "rest" in inst.path
                ]

        _ = [setattr(inst, "patches", []) for inst in instances if path_refresh]
        return instances

    def _checkpostpatch(self, body=None, path=None, patch=False):
        """Make the post file compatible with the system generation

        :param body: contents to be checked
        :type body: str.
        :param path: The URL location to check
        :type path: str.
        :param service: flag to determine if minimum calls should be done.
        :type service: boolean.
        :param url: originating url.
        :type url: str.
        :param sessionid: session id to be used instead of iLO credentials.
        :type sessionid: str.
        :param headers: additional headers to be added to the request.
        :type headers: str.
        :param iloresponse: flag to return the iLO response.
        :type iloresponse: str.
        :param silent: flag to determine if no output should be done.
        :type silent: boolean.
        :param patch: flag to determine if a patch is being made
        :type patch: boolean.
        :returns: modified body and path parameter for target and action respectively
        """
        try:
            if self.typepath.defs.flagforrest:
                if "Target" not in body and not patch:
                    if "/Oem/Hp" in path:
                        body["Target"] = self.typepath.defs.oempath

                if path.startswith("/redfish/v1"):
                    path = path.replace("/redfish", "/rest", 1)

                if "/Actions/" in path:
                    ind = path.find("/Actions/")
                    path = path[:ind]

                if path.endswith("/"):
                    path = path[:-1]
            elif path.startswith("/rest/") and self.typepath.defs.isgen9:
                results = self.get_handler(put_path=path, service=True, silent=True)
                if results and results.status == 200:
                    if results.dict:
                        if "Target" in body:
                            actions = results.dict["Oem"][self.typepath.defs.oemhp]["Actions"]
                        elif "Actions" in body:
                            actions = results.dict["Actions"]
                        else:
                            return path, body

                    allkeys = list(actions.keys())
                    targetkey = [x for x in allkeys if x.endswith(body["Action"])]

                    if targetkey[0].startswith("#"):
                        targetkey[0] = targetkey[0][1:]

                path = path.replace("/rest", "/redfish", 1)
                path = path + "/Actions"

                if "Target" in body:
                    path = path + self.typepath.defs.oempath
                    del body["Target"]

                if targetkey:
                    path = path + "/" + targetkey[0] + "/"

            return path, body
        except Exception as excp:
            raise excp

    def _checkforetagchange(self, instance=None):
        """Function to check the status of the etag

        :param instance: retrieved instance to check etag for change.
        :type instance: dict.
        """
        if instance:
            path = instance.path
            (oldtag, _) = self._gettypeswithetag()
            self._updatemono(path=path, path_refresh=True)
            (newtag, _) = self._gettypeswithetag()
            if (oldtag[path] != newtag[path]) and self.typepath.defs.hpilodatetimetype not in instance.maj_type:
                warning_handler(
                    "The property you are trying to change "
                    "has been updated. Please check entry again "
                    " before manipulating it.\n"
                )
                raise ValueChangedError()

    def _gettypeswithetag(self):
        """Gathers etags of all paths in monolith and their type associations"""
        instancepath = dict()
        instances = dict()

        for inst in self.monolith.iter():
            instancepath[inst.path] = inst.maj_type
            instances[inst.path] = inst.etag

        return [instances, instancepath]



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of gathering the power metrics average on HPE iLO systems
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

#from get_resource_directory import get_resource_directory

def get_powermetrics_average(_redfishobj):

    power_metrics_uri = None

    resource_instances = _redfishobj.get_resource_directory()
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        chassis_uri = _redfishobj.root.obj['Chassis']['@odata.id']
        chassis_response = _redfishobj.get(chassis_uri)
        chassis_members_uri = next(iter(chassis_response.obj['Members']))['@odata.id']
        chassis_members_response = _redfishobj.get(chassis_members_uri)
        power_metrics_uri = chassis_members_response.obj.Oem.Hpe['Power']['@odata.id']
    else:
        for instance in resource_instances:
            #Use Resource directory to find the relevant URI
            if '#Power.' in instance['@odata.type']:
                power_metrics_uri = instance['@odata.id']
                break

    if power_metrics_uri:
        power_metrics_data = _redfishobj.get(power_metrics_uri).obj.Oem.Hpe
        print("\n\nPower Data:\n\n")
        print(json.dumps(power_metrics_data, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = input("Enter iLO IP address: ")
    LOGIN_ACCOUNT = input("Enter Username: ")
    LOGIN_PASSWORD = input("Enter password: ")

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    get_powermetrics_average(REDFISHOBJ)
    REDFISHOBJ.logout()



from setuptools import find_packages, setup

extras = {}

setup(name='python-ilorest-library',
      version='4.7.0.0',
      description='iLO Rest Python Library',
      author='Hewlett Packard Enterprise',
      author_email='rajeevalochana.kallur@hpe.com',
      extras_require=extras,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Topic :: Communications'
      ],
      keywords='Hewlett Packard Enterprise',
      url='https://github.com/HewlettPackard/python-ilorest-library',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      install_requires=[
          'jsonpatch',
          'jsonpath_rw',
          'jsonpointer',
          'urllib3',
          'six'
      ])



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Containers used for REST requests and responses."""
import json
import sys

try:
    from collections import OrderedDict
except ImportError:
    from collections.abc import OrderedDict

from six import BytesIO, StringIO, string_types, text_type
from six.moves import http_client
from urllib3.util import Timeout, Retry


class JSONEncoder(json.JSONEncoder):
    """JSON Encoder class"""

    def default(self, obj):
        """Set defaults in JSON encoder class

        :param obj: object to be encoded into JSON.
        :type obj: RestResponse
        :returns: A JSON :class:`OrderedDict`
        """
        if isinstance(obj, RestResponse):
            jsondict = OrderedDict()
            jsondict["Status"] = obj.status
            jsondict["Headers"] = obj.getheaders()

            if obj.read:
                jsondict["Content"] = obj.dict

            return jsondict
        if isinstance(obj, bytes):
            obj = obj.decode("utf-8")
        if isinstance(obj, Timeout):
            jsondict = OrderedDict()
            jsondict["connect_timeout"] = obj.connect_timeout
            jsondict["read_timeout"] = obj.read_timeout
            return jsondict
        if isinstance(obj, Retry):
            jsondict = OrderedDict()
            jsondict["connect"] = obj.connect
            jsondict["read"] = obj.read
            jsondict["redirect"] = obj.redirect
            return jsondict
        return json.JSONEncoder.default(self, obj)


class JSONDecoder(json.JSONDecoder):
    """Custom JSONDecoder that understands our types"""

    def decode(self, json_string):
        """Decode JSON string

        :param json_string: The JSON string to be decoded into usable data.
        :type json_string: str
        :returns: returns a parsed dict
        """
        parsed_dict = super(JSONDecoder, self).decode(json_string)
        return parsed_dict


class _FakeSocket(BytesIO):
    """slick way to parse a http response.
    http://pythonwise.blogspot.com/2010/02/parse-http-response.html"""

    def makefile(self, *args, **kwargs):
        """Return self object"""
        return self


class RisObject(dict):
    """Converts a JSON/Rest dict into a object so you can use .property notation

    :param d: dictionary to be converted
    :type d: dict
    """

    __getattr__ = dict.__getitem__

    def __init__(self, d):
        """Initialize RisObject"""
        super(RisObject, self).__init__()
        self.update(**dict((k, self.parse(value)) for k, value in list(d.items())))

    @classmethod
    def parse(cls, value):
        """Parse for RIS value

        :param cls: class referenced from class method
        :type cls: RisObject
        :param value: value to be parsed
        :type value: data type
        :returns: returns parsed value
        """
        if isinstance(value, dict):
            return cls(value)
        elif isinstance(value, list):
            return [cls.parse(i) for i in value]

        return value


class RestRequest(object):
    """Holder for Request information

    :param path: The URI path.
    :type path: str
    :param method: method to be implemented
    :type method: str
    :param data: body payload for the rest call
    :type data: dict
    """

    def __init__(self, path, method="GET", data="", url=None):
        self._path = path
        self._body = data
        self._method = method
        self.url = url

    @property
    def path(self):
        """The path the request is made against."""
        return self._path

    @property
    def method(self):
        """The method to implement."""
        return self._method

    @property
    def body(self):
        """The body to pass along with the request, if any."""
        return self._body

    def __str__(self):
        """Format string"""
        body = "" if not self._body else self._body
        try:
            return "{} {}\n\n{}".format(self.method, self.path, body)
        except:
            return "{} {}\n\n{}".format(self.method, self.path, "")


class RestResponse(object):
    """Returned by Rest requests

    :param rest_request: Holder for request information
    :type rest_request: :class:`RestRequest` object
    :param http_response: Response from HTTP
    :type http_response: :class:`HTTPResponse` object
    """

    def __init__(self, rest_request, http_response):
        self._read = None
        self._status = None
        self._headers = None
        self._session_key = None
        self._session_location = None
        self._rest_request = rest_request
        self._http_response = http_response
        self._read = self._http_response.data if http_response is not None else None
        self._ori = self._read

    @property
    def read(self):
        """The response body, attempted to be translated into json, else is a string."""
        if self._read and not isinstance(self._read, text_type):
            self._read = self._read.decode("utf-8", "ignore")
        return self._read

    @read.setter
    def read(self, read):
        """Property for setting _read

        :param read: The data to set to read.
        :type read: str
        """
        if read is not None:
            if isinstance(read, dict):
                read = json.dumps(read, indent=4)
            self._read = read

    def getheaders(self):
        """Get all headers included in the response."""
        return dict(self._http_response.headers) if self._http_response is not None else self._headers

    def getheader(self, name):
        """Case-insensitive search for an individual header

        :param name: The header name to retrieve.
        :type name: str
        :returns: returns a header from HTTP response or None if not found.
        """

        def search_dict(search_key, dct):
            for key, val in dct.items():
                if key.lower() == search_key.lower():
                    return val
            return None

        if self._http_response:
            return search_dict(name, self._http_response.headers)
        return search_dict(name, self._headers)

    def loaddict(self, newdict):
        """Property for setting JSON data. Used during initialization.

        :param newdict: The string data to set as JSON data.
        :type newdict: str
        """
        self._read = json.dumps(newdict, indent=4)

    @property
    def dict(self):
        """The response body data as an dict"""
        try:
            return json.loads(self.read)
        except ValueError as exp:
            if self.path != "/smbios" and self.path != "/cgi-bin/uploadFile":
                sys.stderr.write("An invalid response body was returned: %s" % exp)
            return None

    @property
    def obj(self):
        """The response body data as an object"""
        return RisObject.parse(self.dict)

    @property
    def ori(self):
        """The original response body data"""
        return self._ori

    @property
    def status(self):
        """The status code of the request."""
        if self._status:
            return self._status

        return self._http_response.status if self._http_response is not None else self._status

    @property
    def session_key(self):
        """The saved session key for the connection."""
        if self._session_key:
            return self._session_key

        self._session_key = self.getheader("x-auth-token")
        return self._session_key

    @property
    def session_location(self):
        """The saved session location, used for logging out."""
        if self._session_location:
            return self._session_location

        self._session_location = self.getheader("location")
        return self._session_location

    @property
    def request(self):
        """The saved http request the response was generated by."""
        return self._rest_request

    @property
    def path(self):
        """The path the request was made against."""
        return self.request.path

    def __str__(self):
        """Class string formatter"""
        headerstr = ""
        for kiy, val in self.getheaders().items():
            headerstr += "%s %s\n" % (kiy, val)

        return "%(status)s\n%(headerstr)s\n\n%(body)s" % {
            "status": self.status,
            "headerstr": headerstr,
            "body": self.read,
        }


class RisRestResponse(RestResponse):
    """Returned by Rest requests from CHIF

    :param rest_request: Holder for request information
    :type rest_request: :class:`RestRequest` object
    :param resp_text: text from response to be buffered and read
    :type resp_text: str
    """

    def __init__(self, rest_request, resp_txt):
        """Initialization of RisRestResponse"""
        if not isinstance(resp_txt, string_types):
            resp_txt = "".join(map(chr, resp_txt))
        self._respfh = StringIO(resp_txt)
        self._socket = _FakeSocket(bytearray(list(map(ord, self._respfh.read()))))
        #self._respfh = BytesIO(resp_txt.encode('utf-8'))
        #self._socket = _FakeSocket(bytearray(self._respfh.read()))

        response = http_client.HTTPResponse(self._socket)
        #response = response.decode('utf-8')
        response.begin()
        response.data = response.read()
        response.headers = {ki[0]: ki[1] for ki in response.getheaders()}
        super(RisRestResponse, self).__init__(rest_request, response)


class StaticRestResponse(RestResponse):
    """A RestResponse object used when data is being cached."""

    def __init__(self, **kwargs):
        restreq = None

        if "restreq" in kwargs:
            restreq = kwargs["restreq"]

        super(StaticRestResponse, self).__init__(restreq, None)

        if "Status" in kwargs:
            self._status = kwargs["Status"]

        if "Headers" in kwargs:
            self._headers = kwargs["Headers"]

        if "session_key" in kwargs:
            self._session_key = kwargs["session_key"]

        if "session_location" in kwargs:
            self._session_location = kwargs["session_location"]

        if "Content" in kwargs:
            content = kwargs["Content"]

            if isinstance(content, string_types):
                self._read = content
            else:
                self._read = json.dumps(content)
        else:
            self._read = ""

    def getheaders(self):
        """Function for accessing the headers"""
        returnlist = {}

        if isinstance(self._headers, dict):
            returnlist = self._headers
        elif isinstance(self._headers, (list, tuple)):
            returnlist = {ki[0]: ki[1] for ki in self._headers}
        else:
            for item in self._headers:
                returnlist.update(item.items()[0])

        return returnlist



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def get_EncryptionSettings(restobj):
    resource_instances = get_resource_directory(restobj)
    if resource_instances:
        #Get URI from resource directory
        for instance in resource_instances:
            if "Collection.1.0.0" in instance.Type:
                if "HpSmartStorageArrayController." in instance.MemberType:
                    array_controllers = instance.href
                    break

    types = ["Name", "Model", "SerialNumber", "EncryptionBootPasswordSet",\
             "EncryptionCryptoOfficerPasswordSet",\
             "EncryptionLocalKeyCacheEnabled", "EncryptionMixedVolumesEnabled",\
             "EncryptionPhysicalDriveCount", "EncryptionRecoveryParamsSet",\
             "EncryptionStandaloneModeEnabled", "EncryptionUserPasswordSet"]

    collection = restobj.get(array_controllers)
    for instance in collection.obj.links.Member:
        response = restobj.get(instance.href)

        for item in types:
            sys.stdout.write("\tID:  " + str(response.dict["@odata.id"]) + "\n")
            if item in response.dict:
                sys.stdout.write("\t" + item + ": " + str(response.dict[item]) + "\n")
            else:
                sys.stderr.write("\t" + item + "is not " \
                        "available on HpSmartStorageArrayController resource\n")
        sys.stdout.write("%s" % response)
if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT =  "admin"
    LOGIN_PASSWORD =  "password"

    # Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    REST_OBJ.login()
    sys.stdout.write("\nEXAMPLE 43: Dump EncryptionSettings\n")
    get_EncryptionSettings(REST_OBJ)
    REST_OBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of gathering the Logical Drives on an HPE iLO system
"""

import sys
import json
import argparse
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError
from ilorest_util import get_resource_directory
from ilorest_util import get_gen

def get_SmartArray_LogicalDrives(_redfishobj):

    smartstorage_response = []
    smartarraycontrollers = dict()

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        systems_uri = _redfishobj.root.obj['Systems']['@odata.id']
        systems_response = _redfishobj.get(systems_uri)
        systems_members_uri = next(iter(systems_response.obj['Members']))['@odata.id']
        systems_members_response = _redfishobj.get(systems_members_uri)
        smart_storage_uri = systems_members_response.obj.Oem.Hpe.Links\
                                                                ['SmartStorage']['@odata.id']
        smart_storage_arraycontrollers_uri = _redfishobj.get(smart_storage_uri).obj.Links\
                                                                ['ArrayControllers']['@odata.id']
        smartstorage_response = _redfishobj.get(smart_storage_arraycontrollers_uri).obj['Members']
    else:
        for instance in resource_instances:
            #Use Resource directory to find the relevant URI
            if '#HpeSmartStorageArrayControllerCollection.' in instance['@odata.type']:
                smartstorage_uri = instance['@odata.id']
                print (smartstorage_uri)
                smartstorage_response = _redfishobj.get(smartstorage_uri).obj['Members']
                break

    for controller in smartstorage_response:
        smartarraycontrollers[controller['@odata.id']] = _redfishobj.get(controller['@odata.id']).obj
        sys.stdout.write("Logical Drive URIs for Smart Storage Array Controller \'%s\' : \n" \
                                        % smartarraycontrollers[controller['@odata.id']].get('Id'))
        logicaldrives_uri = smartarraycontrollers[controller['@odata.id']].Links\
                                                                    ['LogicalDrives']['@odata.id']
        logicaldrives_resp = _redfishobj.get(logicaldrives_uri)
        smartarraycontrollers[controller['@odata.id']]['LogicalDrives'] = logicaldrives_resp.dict\
                                                                                        ['Members']
        if not logicaldrives_resp.dict['Members']:
            sys.stderr.write("\tLogical drives are not available for this controller.\n")
        for lds in logicaldrives_resp.dict['Members']:
            sys.stdout.write("\t An associated logical drive: %s\n" % logicaldrives_resp.dict['Name'])
            drive_data = _redfishobj.get(lds['@odata.id']).dict
            print(json.dumps(drive_data, indent=4, sort_keys=True))

def get_SmartArray_LogicalDrives_gen9(_redfishobj):

    smartstorage_response = []
    smartarraycontrollers = dict()

    smartstorage_uri = "/redfish/v1/Systems/1/SmartStorage/ArrayControllers/"
    print (smartstorage_uri)
    smartstorage_response = _redfishobj.get(smartstorage_uri).obj['Members']

    for controller in smartstorage_response:
        smartarraycontrollers[controller['@odata.id']] = _redfishobj.get(controller['@odata.id']).obj
        sys.stdout.write("Logical Drive URIs for Smart Storage Array Controller \'%s\' : \n" \
                                        % smartarraycontrollers[controller['@odata.id']].get('Id'))
        logicaldrives_uri = smartarraycontrollers[controller['@odata.id']].Links\
                                                                    ['LogicalDrives']['@odata.id']
        print (logicaldrives_uri)
        logicaldrives_resp = _redfishobj.get(logicaldrives_uri)
        if logicaldrives_resp.dict['Members@odata.count'] == 0:
            sys.stderr.write("\tLogical drives are not available for this controller.\n")
        else:
            for lds in logicaldrives_resp.dict['Members']:
                sys.stdout.write("\t An associated logical drive: %s\n" % logicaldrives_resp.dict['Name'])
                drive_data = _redfishobj.get(lds['@odata.id']).dict
                print(json.dumps(drive_data, indent=4, sort_keys=True))

if __name__ == "__main__":

    # Initialize parser
    parser = argparse.ArgumentParser(description = "Script to upload and flash NVMe FW")

    parser.add_argument(
        '-i',
        '--ilo',
        dest='ilo_ip',
        action="store",
        help="iLO IP of the server",
        default=None)
    parser.add_argument(
        '-u',
        '--user',
        dest='ilo_user',
        action="store",
        help="iLO username to login",
        default=None)
    parser.add_argument(
        '-p',
        '--password',
        dest='ilo_pass',
        action="store",
        help="iLO password to log in.",
        default=None)

    options = parser.parse_args()

    system_url = "https://" + options.ilo_ip
    print (system_url)

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        redfish_obj = RedfishClient(base_url=system_url, username=options.ilo_user, password=options.ilo_pass)
        # Login with the Redfish client
        redfish_obj.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()
    (ilogen,_) = get_gen(redfish_obj)
    print ("Generation is ", ilogen)
    if int(ilogen) == 5:
        get_SmartArray_LogicalDrives(redfish_obj)
    else:
        get_SmartArray_LogicalDrives_gen9(redfish_obj)
    redfish_obj.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of retrieving bios settings
"""

import sys
import json
import argparse
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError
global DISABLE_RESOURCE_DIR
#from ilorest_util import get_resource_directory
#from ilorest_util import get_gen

def get_bios_setting(_redfishobj):

    bios_uri = None
    bios_data = None
    resource_instances = _redfishobj.get_resource_directory()
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        systems_uri = _redfishobj.root.obj['Systems']['@odata.id']
        systems_response = _redfishobj.get(systems_uri)
        systems_members_uri = next(iter(systems_response.obj['Members']))['@odata.id']
        systems_members_response = _redfishobj.get(systems_members_uri)
        bios_uri = systems_members_response.obj['Bios']['@odata.id']
        bios_data = _redfishobj.get(bios_uri)
    else:
        #Use Resource directory to find the relevant URI
        for instance in resource_instances:
            if '#Bios.' in instance['@odata.type']:
                bios_uri = instance['@odata.id']
                bios_data = _redfishobj.get(bios_uri)
                break

    if bios_data:
        print("\n\nShowing bios attributes before changes:\n\n")
        print(json.dumps(bios_data.dict, indent=4, sort_keys=True))

def get_bios_setting_gen9(_redfishobj):

    bios_uri = "/redfish/v1/systems/1/bios"
    bios_data = None
    #Use Resource directory to find the relevant URI
    bios_data = _redfishobj.get(bios_uri)

    if bios_data:
        print("\n\nShowing bios attributes before changes:\n\n")
        print(json.dumps(bios_data.dict, indent=4, sort_keys=True))


if __name__ == "__main__":

    SYSTEM_URL = input("Enter iLO IP Address: ")
    LOGIN_ACCOUNT = input("Enter Username: ")
    LOGIN_PASSWORD = input("Enter password: ")

#     Initialize parser
##    parser = argparse.ArgumentParser(description = "Script to upload and flash NVMe FW")
##
##    parser.add_argument(
##        '-i',
##        '--ilo',
##        dest='ilo_ip',
##        action="store",
##        help="iLO IP of the server",
##        default=None)
##    parser.add_argument(
##        '-u',
##        '--user',
##        dest='ilo_user',
##        action="store",
##        help="iLO username to login",
##        default=None)
##    parser.add_argument(
##        '-p',
##        '--password',
##        dest='ilo_pass',
##        action="store",
##        help="iLO password to log in.",
##        default=None)
##
##    options = parser.parse_args()
##
##    system_url = "https://" + options.ilo_ip
##    print (system_url)

    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        #redfish_obj = RedfishClient(base_url=system_url, username=options.ilo_user, password=options.ilo_pass)
        redfish_obj = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        redfish_obj.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()
    (ilogen,_) = redfish_obj.get_gen()
    print ("Generation is ", ilogen)
    if int(ilogen) == 5:
        get_bios_setting(redfish_obj)
    else:
        get_bios_setting_gen9(redfish_obj)
    redfish_obj.logout()



 # Copyright 2019 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of uploading a component to the iLO Repository for flashing
"""

import os
import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def upload_firmware(_redfishobj, firmware_loc, compsig_loc, update_repo=True, update_target=False):
    resource_instances = get_resource_directory(_redfishobj)

    if DISABLE_RESOURCE_DIR or not resource_instances:
        #resource directory is not available so we will navigate through paths manually
        update_service_uri = _redfishobj.root.obj['UpdateService']['@odata.id']
    else:
        #obtain all account instances from resource directory
        for instance in resource_instances:
            if '#UpdateService.' in instance['@odata.type']:
                update_service_uri = instance['@odata.id']

    update_service_response = _redfishobj.get(update_service_uri)

    path = update_service_response.obj.HttpPushUri

    body = []
    json_data = {'UpdateRepository': update_repo, 'UpdateTarget': update_target, 'ETag': 'atag', 'Section': 0}
    session_key = _redfishobj.session_key

    filename = os.path.basename(firmware_loc)
    with open(firmware_loc, 'rb') as fle:
        filenameOutput = fle.read()

    filename = os.path.basename(compsig_loc)
    with open(compsig_loc, 'rb') as fle:
        compsigOutput = fle.read()

    session_tuple = ('sessionKey', session_key)
    parameters_tuple = ('parameters', json.dumps(json_data))
    file_tuple = ('file', (filename, filenameOutput, 'application/octet-stream'))
    comp_tuple = ('compsig', (filename, compsigOutput, 'application/octet-stream'))

    #Build the payload from each multipart-form data tuple
    body.append(session_tuple)
    body.append(parameters_tuple)
    body.append(comp_tuple)
    body.append(file_tuple)

    #Create our header dictionary
    header = {'Cookie': 'sessionKey=' + session_key}

    #We pass the whole list payload to post
    resp = _redfishobj.post(path, body, headers=header)

    if resp.status == 400:
        sys.stderr.write("Failed to upload firmware...")
    elif not resp.status in [200, 201]:
        sys.stderr.write("An http response of '%s' was returned.\n" % resp.status)
    else:
        print("Upload complete!\n")

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.0"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.0"
    LOGIN_ACCOUNT = "admin
    LOGIN_PASSWORD = "password

    # The path to the firmware file to upload
    FIRMWARE_PATH = "/path/to/component.exe"
    COMPSIG_PATH = "/path/to/component.compsiq"
    # Upload the firmware file to the iLO Repository
    UPDATE_REPO = True
    # Update the system with the firmware file
    UPDATE_TARGET = False

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = True

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    upload_firmware(REDFISHOBJ, FIRMWARE_PATH, COMPSIG_PATH, UPDATE_REPO, UPDATE_TARGET)

    REDFISHOBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""All Connections for interacting with REST."""
import gzip
import json
import logging
import time

import urllib3
from urllib3 import PoolManager, ProxyManager
from urllib3.exceptions import DecodeError, MaxRetryError

try:
    urllib3.disable_warnings()
    from urllib3.contrib.socks import SOCKSProxyManager
except ImportError:
    pass

import six
from six import BytesIO
from six.moves.urllib.parse import urlencode, urlparse

from redfish.hpilo.risblobstore2 import (
    Blob2OverrideError,
    Blob2SecurityError,
    BlobStore2,
)
from redfish.hpilo.rishpilo import HpIloChifPacketExchangeError
from redfish.rest.containers import RestRequest, RestResponse, RisRestResponse

# ---------End of imports---------


# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------


class RetriesExhaustedError(Exception):
    """Raised when retry attempts have been exhausted."""

    pass


class VnicNotEnabledError(Exception):
    """Raised when retry attempts have been exhausted when VNIC is not enabled."""

    pass


class DecompressResponseError(Exception):
    """Raised when decompressing the response failed."""

    pass


class InvalidCredentialsError(Exception):
    """Raised when invalid credentials have been provided."""

    pass


class InvalidCertificateError(Exception):
    """Raised when a invalid certificate has been provided."""

    pass


class ChifDriverMissingOrNotFound(Exception):
    """Raised when CHIF driver is missing or not found."""

    pass


class SecurityStateError(Exception):
    """Raised when there is a strict security state without authentication."""

    pass


class OneTimePasscodeError(Exception):
    """Raised when OTP is sent to the registered email."""

    pass


class TokenExpiredError(Exception):
    """Raised when OTP entered has expired."""

    pass


class UnauthorizedLoginAttemptError(Exception):
    """Raised when Login is Unauthorized"""

    pass


class HttpConnection(object):
    """HTTP connection capable of authenticating with HTTPS and Http/Socks Proxies

    :param base_url: The URL to make HTTP calls against
    :type base_url: str
    :param \\**client_kwargs: Arguments to pass to the connection initialization. These are"
        "passed to a urllib3 `PoolManager <https://urllib3.readthedocs.io/en/latest/reference/"
        "index.html?highlight=PoolManager#urllib3.PoolManager>`_. All arguments that can be passed to "
        "a PoolManager are valid arguments."
    """

    def __init__(self, base_url, cert_data, **client_kwargs):
        self._conn = None
        self.base_url = base_url
        # Default values for connection properties
        self._connection_properties = {
            'timeout': urllib3.util.Timeout(connect=4800, read=4800),
            'retries': urllib3.util.Retry(connect=50, read=50, redirect=50),
        }
        self._connection_properties.update(client_kwargs)
        if cert_data:
            if ("cert_file" in cert_data and cert_data["cert_file"]) or (
                "ca_certs" in cert_data and cert_data["ca_certs"]
            ):
                self._connection_properties.update({"ca_cert_data": cert_data})
        self._proxy = self._connection_properties.pop("proxy", None)
        self.session_key = self._connection_properties.pop("session_key", None)
        self.log_dir = self._connection_properties.pop("log_dir", None)
        self._init_connection()

    @property
    def proxy(self):
        """The proxy, if any."""
        return self._proxy

    @proxy.setter
    def proxy(self, proxy):
        """set the proxy"""
        self._proxy = proxy

    def _init_connection(self):
        """Function for initiating connection with remote server"""
        cert_reqs = "CERT_NONE"
        if self._connection_properties.get("ca_cert_data"):
            LOGGER.info("Using CA cert to confirm identity.")
            cert_reqs = "CERT_NONE"
            self._connection_properties.update(self._connection_properties.pop("ca_cert_data"))

        if self.proxy:
            if self.proxy.startswith("socks"):
                LOGGER.info("Initializing a SOCKS proxy.")
                http = SOCKSProxyManager(
                    self.proxy,
                    cert_reqs=cert_reqs,
                    maxsize=50,
                    **self._connection_properties
                )
            else:
                LOGGER.info("Initializing a HTTP proxy.")
                http = ProxyManager(
                    self.proxy,
                    cert_reqs=cert_reqs,
                    maxsize=50,
                    **self._connection_properties
                )
        else:
            LOGGER.info("Initializing no proxy.")
            try:
                self._connection_properties.pop("ca_cert_data")
            except KeyError:
                pass

            http = PoolManager(
                cert_reqs=cert_reqs,
                maxsize=50,
                **self._connection_properties
            )

        self._conn = http.request

    def rest_request(self, path, method="GET", args=None, body=None, headers=None):
        """Format and do HTTP Rest request

        :param path: The URI path to perform the operation on.
        :type path: str
        :param method: method to perform on the path.
        :type method: str
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :param body: body payload to include in the request if needed.
        :type body: dict
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        # TODO: Need to remove redfish.dmtf.org calls from here, add to their own HttpConnection
        files = None
        request_args = {}
        if isinstance(path, bytes):
            path = path.decode("utf-8")
            external_uri = True if "redfish.dmtf.org" in path else False
        else:
            external_uri = True if "redfish.dmtf.org" in path else False
        headers = {} if external_uri else headers
        reqpath = path.replace("//", "/") if not external_uri else path

        if body is not None:
            if body and isinstance(body, list) and isinstance(body[0], tuple):
                files = body
                body = None
            elif isinstance(body, (dict, list)):
                headers["Content-Type"] = "application/json"
                body = json.dumps(body)
            elif not files:
                headers["Content-Type"] = "application/octet-stream"

            if method == "PUT":
                resp = self.rest_request(method="HEAD", path=path, headers=headers)

                try:
                    if resp.getheader("content-encoding") == "gzip":
                        buf = BytesIO()
                        gfile = gzip.GzipFile(mode="wb", fileobj=buf)

                        try:
                            gfile.write(str(body).encode("utf-8") if six.PY3 else str(body))
                        finally:
                            gfile.close()

                        compresseddata = buf.getvalue()
                        if compresseddata:
                            data = bytearray()
                            data.extend(memoryview(compresseddata))
                            body = data
                except BaseException as excp:
                    LOGGER.error("Error occur while compressing body: %s", excp)
                    raise

        if args:
            if method == "GET":
                reqpath += "?" + urlencode(args)
            elif method == "PUT" or method == "POST" or method == "PATCH":
                headers["Content-Type"] = "application/x-www-form-urlencoded"
                body = urlencode(args)

        # TODO: ADD to the default headers?
        if headers is not None:
            headers["Accept-Encoding"] = "gzip"
        restreq = RestRequest(path, method, data=files if files else body, url=self.base_url)

        if LOGGER.isEnabledFor(logging.DEBUG):
            try:
                logbody = None
                if restreq.body:
                    if restreq.body[0] == "{":
                        logbody = restreq.body
                    else:
                        raise KeyError()
                if restreq.method in ["POST", "PATCH"]:
                    debugjson = json.loads(restreq.body)
                    if "Password" in debugjson.keys():
                        debugjson["Password"] = "******"
                    if "OldPassword" in debugjson.keys():
                        debugjson["OldPassword"] = "******"
                    if "NewPassword" in debugjson.keys():
                        debugjson["NewPassword"] = "******"
                    logbody = json.dumps(debugjson)
                    logbody = logbody.replace("\\\\", "\\")
                LOGGER.debug(
                    "HTTP REQUEST: %s\n\tPATH: %s\n\t" "HEADERS: %s\n\tBODY: %s",
                    restreq.method,
                    restreq.path,
                    headers,
                    logbody,
                )
            except:
                LOGGER.debug(
                    "HTTP REQUEST: %s\n\tPATH: %s\n\tBODY: %s",
                    restreq.method,
                    restreq.path,
                    "binary body",
                )

        inittime = time.time()
        reqfullpath = self.base_url + reqpath if not external_uri else reqpath

        # To ensure we don't have unicode/string merging issues in httplib of Python 2
        if isinstance(reqfullpath, six.text_type):
            reqfullpath = str(reqfullpath)

        if headers:
            request_args["headers"] = headers
        if files:
            request_args["fields"] = files
        else:
            request_args["body"] = body
        try:
            resp = self._conn(method, reqfullpath, **request_args)
        except MaxRetryError:
            vnic_url = "16.1.15.1"
            if reqfullpath.find(vnic_url) != -1:
                raise VnicNotEnabledError()
            raise RetriesExhaustedError()
        except DecodeError:
            raise DecompressResponseError()

        endtime = time.time()
        LOGGER.info("Response Time to %s: %s seconds.", restreq.path, str(endtime - inittime))

        restresp = RestResponse(restreq, resp)
        #        if restresp.request.body:
        #            respbody = restresp.read
        #            respbody = respbody.replace("\\\\", "\\")

        if LOGGER.isEnabledFor(logging.DEBUG):
            headerstr = ""
            if restresp is not None:
                respheader = restresp.getheaders()
                for kiy, headerval in respheader.items():
                    headerstr += "\t" + kiy + ": " + headerval + "\n"
                try:
                    LOGGER.debug(
                        "HTTP RESPONSE for %s:\nCode:%s\nHeaders:" "\n%s\nBody Response of %s: %s",
                        restresp.request.path,
                        str(restresp._http_response.status) + " " + restresp._http_response.reason,
                        headerstr,
                        restresp.request.path,
                        restresp.read,
                    )
                except:
                    LOGGER.debug("HTTP RESPONSE:\nCode:%s", restresp)
            else:
                LOGGER.debug("HTTP RESPONSE: No HTTP Response obtained")

        return restresp

    def cert_login(self):
        """Login using a certificate."""
        resp = self.rest_request("/html/login_cert.html", "GET")
        if resp.status == 200 or resp.status == 201:
            token = resp.getheader("X-Auth-Token")
            location = resp.getheader("Location")
        else:
            raise InvalidCertificateError("")

        return token, location


class Blobstore2Connection(object):
    """A connection for communicating locally with HPE servers

    :param \\**conn_kwargs: Arguments to pass to the connection initialization.

    Possible arguments for *\\**conn_kwargs* include:

    :username: The username to login with
    :password: The password to login with
    """

    _http_vsn_str = "HTTP/1.1"
    blobstore_headers = {"Accept": "*/*", "Connection": "Keep-Alive"}

    def __del__(self):
        """Clear channel"""
        self._conn = None

    def __init__(self, **conn_kwargs):
        self._conn = None
        self.base_url = "blobstore://."
        self._connection_properties = conn_kwargs
        self.session_key = self._connection_properties.pop("sessionid", None)
        self._init_connection(**self._connection_properties)

    def _init_connection(self, **kwargs):
        """Initiate blobstore connection"""
        # mixed security modes require a password at all times
        username = kwargs.pop("username", "nousername")
        if isinstance(username, bytes):
            username = username.decode("utf-8")
        password = kwargs.pop("password", "nopassword")
        if isinstance(password, bytes):
            password = password.decode("utf-8")
        log_dir = kwargs.pop("log_dir", "")
        try:
            correctcreds = BlobStore2.initializecreds(username=username, password=password, log_dir=log_dir)
            bs2 = BlobStore2()
            if not correctcreds:
                security_state = int(bs2.get_security_state())
                if security_state == 3:
                    raise InvalidCredentialsError(0)
                else:
                    raise SecurityStateError(security_state)
        except Blob2SecurityError:
            raise InvalidCredentialsError(0)
        except HpIloChifPacketExchangeError as excp:
            LOGGER.info("Exception: %s", str(excp))
            raise ChifDriverMissingOrNotFound()
        except Exception as excp:
            if str(excp) == "chif":
                raise ChifDriverMissingOrNotFound()
            else:
                raise
        else:
            self._conn = bs2

    def rest_request(self, path="", method="GET", args=None, body=None, headers=None):
        """Rest request for blobstore client

        :param path: The URI path to perform the operation on.
        :type path: str
        :param method: method to perform on the path.
        :type method: str
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :param body: body payload to include in the request if needed.
        :type body: dict
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        # default headers if not passed in - otherwise will throw on .update call
        if headers is None:
            headers = {}
        else:
            headers.update(Blobstore2Connection.blobstore_headers)
        if isinstance(path, bytes):
            path = path.decode("utf-8")
        reqpath = path.replace("//", "/")

        oribody = body
        if body is not None:
            if isinstance(body, (dict, list)):
                headers["Content-Type"] = "application/json"
                if isinstance(body, bytes):
                    body = body.decode("utf-8")
                body = json.dumps(body)
            elif isinstance(body, bytes):
                headers["Content-Type"] = "application/octet-stream"
                body = bytearray(body)
            else:
                headers["Content-Type"] = "application/x-www-form-urlencoded"
                body = urlencode(body)

            if method == "PUT":
                resp = self.rest_request(path=path, headers=headers)

                try:
                    if resp.getheader("content-encoding") == "gzip":
                        buf = BytesIO()
                        gfile = gzip.GzipFile(mode="wb", fileobj=buf)

                        try:
                            gfile.write(str(body).encode("utf-8") if six.PY3 else str(body))
                        finally:
                            gfile.close()

                        compresseddata = buf.getvalue()
                        if compresseddata:
                            data = bytearray()
                            data.extend(memoryview(compresseddata))
                            body = data
                except BaseException as excp:
                    LOGGER.error("Error occur while compressing body: %s", excp)
                    raise

            headers["Content-Length"] = len(body)

        if args:
            if method == "GET":
                reqpath += "?" + urlencode(args)
            elif method == "PUT" or method == "POST" or method == "PATCH":
                headers["Content-Type"] = "application/x-www-form-urlencoded"
                body = urlencode(args)

        str1 = "{} {} {}\r\n".format(method, reqpath, Blobstore2Connection._http_vsn_str)
        str1 += "Host: \r\n"
        str1 += "Accept-Encoding: gzip\r\n"
        for header, value in headers.items():
            str1 += "{}: {}\r\n".format(header, value)

        str1 += "\r\n"

        if body and len(body) > 0:
            if isinstance(body, bytearray):
                str1 = bytearray(str1.encode("utf-8")) + body
            else:
                #if isinstance(body, bytes):
                #    body = body.decode("utf-8")
                str1 += body

        if not isinstance(str1, bytearray):
            str1 = bytearray(str1.encode("utf-8"))

        if LOGGER.isEnabledFor(logging.DEBUG):
            try:
                logbody = None
                if body:
                    if body[0] == "{":
                        logbody = body
                    else:
                        raise
                if method in ["POST", "PATCH"]:
                    debugjson = json.loads(body)
                    if "Password" in debugjson.keys():
                        debugjson["Password"] = "******"
                    if "OldPassword" in debugjson.keys():
                        debugjson["OldPassword"] = "******"
                    if "NewPassword" in debugjson.keys():
                        debugjson["NewPassword"] = "******"
                    logbody = json.dumps(debugjson)

                LOGGER.debug(
                    "Blobstore REQUEST: %s\n\tPATH: %s\n\tHEADERS: " "%s\n\tBODY: %s",
                    method,
                    str(headers),
                    path,
                    logbody,
                )
            except:
                LOGGER.debug(
                    "Blobstore REQUEST: %s\n\tPATH: %s\n\tHEADERS: " "%s\n\tBODY: %s",
                    method,
                    str(headers),
                    path,
                    "binary body",
                )

        inittime = time.time()

        for idx in range(5):
            try:
                resp_txt = self._conn.rest_immediate(str1)
                break
            except Blob2OverrideError:
                if idx == 4:
                    raise Blob2OverrideError(2)
                else:
                    continue

        endtime = time.time()

        LOGGER.info("iLO Response Time to %s: %s secs.", path, str(endtime - inittime))

        if resp_txt is not None:
            # Dummy response to support a bad host response
            if len(resp_txt) == 0:
                resp_txt = (
                    "HTTP/1.1 500 Not Found\r\nAllow: "
                    "GET\r\nCache-Control: no-cache\r\nContent-length: "
                    "0\r\nContent-type: text/html\r\nDate: Tues, 1 Apr 2025 "
                    "00:00:01 GMT\r\nServer: "
                    "HP-iLO-Server/1.30\r\nX_HP-CHRP-Service-Version: 1.0.3\r\n\r\n\r\n"
                )

            restreq = RestRequest(path, method, data=body, url=self.base_url)
            rest_response = RisRestResponse(restreq, resp_txt)

            if rest_response.status in range(300, 399) and rest_response.status != 304:
                newloc = rest_response.getheader("location")
                newurl = urlparse(newloc)

                rest_response = self.rest_request(newurl.path, method, args, oribody, headers)

            try:
                if rest_response.getheader("content-encoding") == "gzip":
                    if hasattr(gzip, "decompress"):
                        rest_response.read = gzip.decompress(rest_response.ori)
                    else:
                        compressedfile = BytesIO(rest_response.ori)
                        decompressedfile = gzip.GzipFile(fileobj=compressedfile)
                        rest_response.read = decompressedfile.read()
            except Exception:
                pass
            if LOGGER.isEnabledFor(logging.DEBUG):
                headerstr = ""
                headerget = rest_response.getheaders()
                for header in headerget:
                    headerstr += "\t" + header + ": " + headerget[header] + "\n"
                try:
                    LOGGER.debug(
                        "Blobstore RESPONSE for %s:\nCode: %s\nHeaders:" "\n%s\nBody of %s: %s",
                        rest_response.request.path,
                        str(rest_response._http_response.status) + " " + rest_response._http_response.reason,
                        headerstr,
                        rest_response.request.path,
                        rest_response.read,
                    )
                except:
                    LOGGER.debug(
                        "Blobstore RESPONSE for %s:\nCode:%s",
                        rest_response.request.path,
                        rest_response,
                    )
            return rest_response

    def cert_login(self):
        """Login using a certificate."""
        # local cert login is only available on iLO 5
        token = self.cert_login()
        resp = self.rest_request("/redfish/v1/SessionService/Sessions/", "GET")
        if resp.status == 200:
            try:
                location = resp.obj.Oem.Hpe.Links.MySession["@odata.id"]
            except KeyError:
                raise InvalidCertificateError("")
        else:
            raise InvalidCertificateError("")

        return token, location



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of generating a certificate signing request for HPE iLO systems
"""
import sys
import json
import time
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def generate_csr(_redfishobj, csr_file, csr_properties):

    csr_uri = None
    generate_csr_uri = None

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        managers_uri = _redfishobj.root.obj['Managers']['@odata.id']
        managers_response = _redfishobj.get(managers_uri)
        managers_members_uri = next(iter(managers_response.obj['Members']))['@odata.id']
        managers_members_response = _redfishobj.get(managers_members_uri)
        security_service_uri = managers_members_response.obj.Oem.Hpe.Links\
                                                                ['SecurityService']['@odata.id']
        security_service_response = _redfishobj.get(security_service_uri)
        csr_uri = security_service_response.obj.Links['HttpsCert']['@odata.id']
        https_cert_response = _redfishobj.get(csr_uri)
        generate_csr_uri = https_cert_response.obj['Actions']['#HpeHttpsCert.GenerateCSR']\
                                                                                    ['target']
    else:
        #Use Resource directory to find the relevant URI
        for instance in resource_instances:
            if '#HpeHttpsCert.' in instance['@odata.type']:
                csr_uri = instance['@odata.id']
                generate_csr_uri = _redfishobj.get(csr_uri).obj['Actions']\
                                                        ['#HpeHttpsCert.GenerateCSR']['target']
                break

    if generate_csr_uri:
        body = dict()
        body["Action"] = "HpeHttpsCert.GenerateCSR"
        body["City"] = csr_properties["City"]
        body["CommonName"] = csr_properties["CommonName"]
        body["Country"] = csr_properties["Country"]
        body["OrgName"] = csr_properties["OrgName"]
        body["OrgUnit"] = csr_properties["OrgUnit"]
        body["State"] = csr_properties["State"]
        resp = _redfishobj.post(generate_csr_uri, body)
        if resp.status in [200, 201]:
            sys.stdout.write("Generating CSR, this may take a few minutes\n")
            sys.stdout.write("Sleeping for 5 minutes...\n")
            time.sleep(600)
            csr_resp = _redfishobj.get(csr_uri).dict['CertificateSigningRequest']
            with open(csr_file, 'wb') as csroutput:
                csroutput.write(csr_resp)
            sys.stdout.write("CSR Data saved to file: '%s'\n" % csr_file)
        else:
            #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended
            #info error message to see what went wrong
            if resp.status not in [200, 201]:
                try:
                    print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, \
                                                                                    sort_keys=True))
                except Exception as excp:
                    sys.stderr.write("A response error occurred, unable to access iLO "\
                                     "Extended Message Info...\n")

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #CSR Dictionary, each value should be updated to reflect the desired CSR
    CSR_DICT = {"City" : "City", "CommonName": "Common Name", "Country": "US", \
                "OrgName": "Organization", "OrgUnit": "Unit", "State": "State"}

    #After CSR is generated, a file will be created and the CSR will be downloaded
    CSR_FILE = "csr.txt"
    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    generate_csr(REDFISHOBJ, CSR_FILE, CSR_DICT)
    REDFISHOBJ.logout()
